import tensorflow as tf

class Generator(tf.keras.Model):
    def __init__(self, n_input=7, n_past_points=40, n_labels=3, data_noise_dim=50):

        super(Generator, self).__init__()
        # DATA
        self.d1 = tf.keras.layers.Dense(n_past_points * n_input, use_bias=False, input_shape=(data_noise_dim,))
        self.bn1 = tf.keras.layers.BatchNormalization()
        self.lRe1 = tf.keras.layers.LeakyReLU()

        self.conv1 = tf.keras.layers.Conv1DTranspose(32, 5, strides=1, padding='same', use_bias=False)
        self.bn2 = tf.keras.layers.BatchNormalization()
        self.lRe2 = tf.keras.layers.LeakyReLU()

        # LABEL
        self.d12 = tf.keras.layers.Dense(n_past_points * n_labels, use_bias=False, input_shape=(3,))
        self.bn12 = tf.keras.layers.BatchNormalization()
        self.lRe12 = tf.keras.layers.LeakyReLU()

        self.conv12 = tf.keras.layers.Conv1DTranspose(32, 5, strides=1, padding='same', use_bias=False)
        self.bn22 = tf.keras.layers.BatchNormalization()
        self.lRe22 = tf.keras.layers.LeakyReLU()

        # DATA + LABEL
        self.conv2 = tf.keras.layers.Conv1DTranspose(32, 5, strides=2, padding='same', use_bias=False)
        self.bn3 = tf.keras.layers.BatchNormalization()
        self.lRe3 = tf.keras.layers.LeakyReLU()

        self.conv3 = tf.keras.layers.Conv1D(n_input, 5, strides=2, padding='same', use_bias=False, activation='tanh')
        self.bn32 = tf.keras.layers.BatchNormalization()
        self.lRe32 = tf.keras.layers.LeakyReLU()

    def call(self, inputs, training=None):
        # data
        inputs_1 = tf.slice(inputs, [0, 0], [-1, 50])
        # label
        inputs_2 = tf.slice(inputs, [0, 50], [-1, 3])

        # INPUT
        x = self.d1(inputs_1)
        x = self.bn1(x, training=training)
        x = self.lRe1(x, training=training)
        x = tf.reshape(x, [-1, 40, 7])

        # LABELS
        y = self.d12(inputs_2)
        y = self.bn2(y, training=training)
        y = self.lRe2(y, training=training)
        y = tf.reshape(y, [-1, 40, 3])

        # LABELS + INPUT
        x = tf.concat([x, y], axis=2)
        x = self.conv12(x)
        x = self.bn22(x, training=training)
        x = self.lRe22(x, training=training)

        x = self.conv2(x)
        x = self.bn3(x, training=training)
        x = self.lRe3(x, training=training)

        x = self.conv3(x)
        x = self.bn32(x, training=training)
        x = self.lRe32(x, training=training)

        return x

    def loadModel(self, model_path):
        self.load_weights(model_path)

class Model(tf.keras.Model):

    def __init__(self, n_classes, n_input):
        super(Model, self).__init__()

        self.conv1 = tf.keras.layers.Conv1D(filters=32, kernel_size=6, padding='causal')
        self.p1 = tf.keras.layers.MaxPooling1D(pool_size=2, strides=2, padding='same')
        self.avg1 = tf.keras.layers.AveragePooling1D(pool_size=2, strides=2, padding='same')
        self.bn1 = tf.keras.layers.BatchNormalization()
        self.f1 = tf.keras.layers.Flatten()
        self.drop1 = tf.keras.layers.Dropout(rate=0.4)

        self.conv2 = tf.keras.layers.Conv1D(filters=32, kernel_size=6, padding="causal")
        self.p2 = tf.keras.layers.MaxPooling1D(pool_size=2, strides=2, padding='same')
        self.avg2 = tf.keras.layers.AveragePooling1D(pool_size=2, strides=2, padding='same')
        self.bn2 = tf.keras.layers.BatchNormalization()
        self.f2 = tf.keras.layers.Flatten()
        self.drop2 = tf.keras.layers.Dropout(rate=0.4)

        self.d1 = tf.keras.layers.Dense(64, kernel_regularizer='l1', activation='relu')
        self.drop3 = tf.keras.layers.Dropout(rate=0.4)
        self.bn3 = tf.keras.layers.BatchNormalization()

        self.d2 = tf.keras.layers.Dense(32, kernel_regularizer='l1', activation='relu')
        self.drop4 = tf.keras.layers.Dropout(rate=0.4)
        self.bn4 = tf.keras.layers.BatchNormalization()

        self.d3 = tf.keras.layers.Dense(16, kernel_regularizer='l1', activation='relu')
        self.drop5 = tf.keras.layers.Dropout(rate=0.4)
        self.bn5 = tf.keras.layers.BatchNormalization()

        self.d4 = tf.keras.layers.Dense(n_classes, activation='sigmoid')

    def call(self, inputs, training=None):
        x = self.conv1(inputs)
        x = self.avg1(x)
        x = self.bn1(x, training=training)
        x = self.drop1(x, training=training)

        x = self.conv2(x)
        x = self.avg2(x)
        x = self.bn2(x, training=training)
        x = self.f2(x, training=training)
        x = self.drop2(x, training=training)

        x = self.d1(x)
        x = self.drop3(x, training=training)
        x = self.bn3(x, training=training)

        x = self.d2(x)
        x = self.drop4(x, training=training)
        x = self.bn4(x, training=training)

        x = self.d3(x)
        x = self.drop5(x, training=training)
        x = self.bn5(x, training=training)

        return self.d4(x)

    def loadModel(self, model_path):
        # self.model = Model(n_classes, n_input)
        self.load_weights(model_path)

class TCN(tf.keras.Model):
    def __init__(self, num_classes, num_filters, kernel_size, dilation_rates):
        super(TCN, self).__init__()
        self.conv_layers = []

        for dilation_rate in dilation_rates:
            self.conv_layers.append(tf.keras.layers.Conv1D(filters=num_filters,
                                                           kernel_size=kernel_size,
                                                           padding='causal',
                                                           dilation_rate=dilation_rate,
                                                           activation='sigmoid'))
        self.f = tf.keras.layers.Flatten()
        self.drop = tf.keras.layers.Dropout(rate=0.4)
        self.dense = tf.keras.layers.Dense(16, kernel_regularizer='l1', activation='sigmoid')
        self.bn = tf.keras.layers.BatchNormalization()

        self.output_layer = tf.keras.layers.Dense(num_classes, activation='sigmoid')

    def call(self, inputs, training=None):
        x = inputs

        for conv_layer in self.conv_layers:
            x = conv_layer(x)
        x = self.f(x)
        x = self.drop(x)
        x = self.dense(x)
        x = self.bn(x, training=training)
        output = self.output_layer(x)
        return output

    def loadModel(self, model_path):
        # self.model = Model(n_classes, n_input)
        self.load_weights(model_path)


import pandas as pd
import numpy as np
from pandas import read_csv

class Test:

    def __init__(self, filename):
        self.filename = filename

    def set_data_path(self, data_path):
        self.data_path = data_path

    def set_data_extension(self, data_extension):
        self.data_extension = data_extension

    def clear_file_name(self):
        if '.' in self.filename and self.filename.split('.')[-1] == '.csv':
            self.filename_cleared = self.filename.split('.')[0]
        else:
            self.filename_cleared = self.filename

        self.filename_cleared = self.filename_cleared.rpartition('/')[-1]

    def infer_metabolimeter(self, optional_data=[]):

        import chardet

        try:
            if self.data_extension == '.csv':
                df = read_csv(self.filename + self.data_extension)
                print('Just reading a csv file')
            if self.data_extension == '.txt':
                df = read_csv(self.filename + self.data_extension, sep="\t", header=None, skiprows=3)
                print('Just reading a txt file')
                self.metabolimeter = 'vyiare'
            if self.data_extension == '.xlsx' or self.data_extension == '.xls':
                print('Attempting to read an Excel file')
                df = pd.read_excel(self.filename + self.data_extension)
        except:
            try:
                f = open(self.filename + self.data_extension, encoding="utf8", errors="ignore")
                df = read_csv(f, header=0)
            except:
                df = optional_data

        self.df = df

        if 'VO2_I' in df.columns:
            self.metabolimeter = 'generated_pyoxynet'

        if 'Polso O2' in df.columns:
            self.metabolimeter = 'vintus'

        if df.isnull().apply(lambda x: all(x), axis=1)[0]:
            # print('Most probably we have a Cortex file at first sight')
            # print(df.columns[0])
            # print(df)
            self.metabolimeter = 'cortex'

        try:
            if df['Unnamed: 0'][0] == 'CPET Results':
                # print('Most probably we have a Cortex file at second sight')
                # print(df)
                self.metabolimeter = 'cortex'
        except:
            pass

        try:
            if 'Messzeit' in df['ID:'].values:
                # print('Most probably we have a Cortex file at second sight')
                # print(df)
                self.metabolimeter = 'cortex_bruce'
        except:
            pass

        try:
            if 'Patient' in df.columns[0]:
                # print('Most probably we have a Cortex file at second sight')
                # print(df)
                self.metabolimeter = 'cortex_bruce_2'
        except:
            pass

        try:
            if 'Marqueur' in df[df.columns[2]].tolist():
                print('Most probably we have a Cortex file that Bruce provided')
                self.metabolimeter = 'cortex_bruce_3'

            if 'Work' in df.columns[2]:
                print('Most probably we have an Italian file from cardiac patients (CENTRO MONZINO)')
                self.metabolimeter = 'centro-monzino'

            if 'MINUTE' in df.columns[0] or 'NOM :' in df.columns[0]:
                print('Most probably we have a Mourot file from cardiac patients')
                self.metabolimeter = 'mourot_cardiac'

            if 'Time' in df.columns[0] and 'Load' in df.columns[1]:
                print('Most probably we have a Low file')
                self.metabolimeter = 'low'

            if 'Time' in df.columns[0] and 'Time' in df.columns[1]:
                print('Most probably we have a VO2Master file')
                self.metabolimeter = 'VO2Master'

            if 'Zeit' in df.columns[0]:
                print('Most probably we have a Unisbz file')
                self.metabolimeter = 'unisbz'

            if 'Codice' in df.columns[0]:
                print('Most probably we have a Italian Cosmed file')
                self.metabolimeter = 'cosmed'

            if df[df.columns[0]][0] == 'Nom de famille':
                print('Most probably we have a French Cosmed file')
                self.metabolimeter = 'cosmed'

            if df[df.columns[0]][0] == 'Sobrenome':
                print('Most probably we have a Brasilian Cosmed file')
                self.metabolimeter = 'cosmed'

            if 'ID code' in df.columns[0]:
                print('Most probably we have an English Cosmed file')
                self.metabolimeter = 'cosmed'

            if 'Code ID' in df.columns[0]:
                print('Most probably we have an English Cosmed file')
                self.metabolimeter = 'cosmed'

            if 'ID1' in df.columns[0]:
                print('Most probably we have an English Cosmed file')
                self.metabolimeter = 'cosmed'

            if 'Temps' in df[df.columns[0]].unique():
                print('Most probably we have a French Mourot file')
                self.metabolimeter = 'mourot'

            if 'TEMPS' in df.columns[0]:
                print('Most probably we have an French file')
                self.metabolimeter = 'mourot_COPD'

            if 'RF' in df.columns and 't' in df.columns and 'VE/VO2' in df.columns and 'VE/VCO2' in df.columns and 'R' in df.columns:
                print('Most probably we have an file coming from the Exercise Threshold App')
                self.metabolimeter = 'exercise_threshold_app'

            if " V'O2   " in df.columns and " V'CO2  " in df.columns:
                print('Most probably we have an file coming from the Uni of Ess')
                self.metabolimeter = 'uni_of_essex'

        except:
            pass

    def load_file(self):

        from . import utilities
        df = self.df

        if self.metabolimeter == 'vintus':

            starting_index = df.loc[df['Tempo'] == 'min'].index
            ref_string = 'Tempo'
            ventilatory_data = df[starting_index[0]+1:-1]

            # print('Weight')
            self.weight = 70
            # print('Height')
            self.height = 180
            # print('Age')
            self.age = 40
            # print('Gender')
            self.gender = 'M'

            # print('Reading data')
            n_rows = len(ventilatory_data.index)
            # initialise variables
            self.time = np.zeros((n_rows,), dtype=np.float32)
            self.VO2 = np.zeros((n_rows,), dtype=np.float32)
            self.VCO2 = np.zeros((n_rows,), dtype=np.float32)
            self.HR = np.zeros((n_rows,), dtype=np.float32)
            self.Rf = np.zeros((n_rows,), dtype=np.float32)
            self.VE = np.zeros((n_rows,), dtype=np.float32)
            self.PetO2 = np.zeros((n_rows,), dtype=np.float32)
            self.PetCO2 = np.zeros((n_rows,), dtype=np.float32)
            self.load = np.zeros((n_rows,), dtype=np.float32)

            i = 0
            for time_sec in ventilatory_data[ventilatory_data.columns[0]].values:
                # print(i)
                try:
                    self.time[i] = utilities.get_sec(time_sec)
                    self.VO2[i] = float(ventilatory_data[ventilatory_data.columns[5]].values[i])
                    self.VCO2[i] = float(ventilatory_data[ventilatory_data.columns[6]].values[i])
                    self.HR[i] = float(ventilatory_data[ventilatory_data.columns[2]].values[i])

                    try:
                        self.load[i] = float(ventilatory_data[ventilatory_data.columns[1]].values[i])
                    except:
                        self.load[i] = 0

                    self.VE[i] = float(ventilatory_data[ventilatory_data.columns[3]].values[i])
                    self.Rf[i] = 0
                    self.PetO2[i] = float(ventilatory_data[ventilatory_data.columns[15]].values[i]) * 7.50062
                    self.PetCO2[i] = float(ventilatory_data[ventilatory_data.columns[14]].values[i]) * 7.50062
                    i += 1
                except:
                    pass

        if self.metabolimeter == 'vyiare':

            with open(self.filename + self.data_extension) as f:
                line1, line2 = next(f), next(f)

            # print('Weight')
            try:
                self.weight = int(line2[line2.find('Weight')+8:line2.find('Weight')+10])
                # print('Height')
                self.height = int(line1[line1.find('Height')+8:line1.find('Height')+11])
                # print('Age')
                self.age = 40
                # print('Gender')
                self.gender = line1[line1.find('Gender')+8].capitalize()
            except:
                self.weight = 70
                # print('Height')
                self.height = 170
                # print('Age')
                self.age = 40
                # print('Gender')
                self.gender = 'M'

            # print('Reading data')
            n_rows = len(df.index)
            # initialise variables
            self.time = np.zeros((n_rows - 3,), dtype=np.float32)
            self.VO2 = np.zeros((n_rows - 3,), dtype=np.float32)
            self.VCO2 = np.zeros((n_rows - 3,), dtype=np.float32)
            self.HR = np.zeros((n_rows - 3,), dtype=np.float32)
            self.Rf = np.zeros((n_rows - 3,), dtype=np.float32)
            self.VE = np.zeros((n_rows - 3,), dtype=np.float32)
            self.PetO2 = np.zeros((n_rows - 3,), dtype=np.float32)
            self.PetCO2 = np.zeros((n_rows - 3,), dtype=np.float32)
            self.load = np.zeros((n_rows - 3,), dtype=np.float32)

            i = 0
            for time_sec in df[0].values[2:-1]:
                # print(i)
                try:
                    self.time[i] = utilities.get_sec(time_sec)
                    self.VO2[i] = float(df[df.columns[4]].values[i+2])
                    self.VCO2[i] = float(df[df.columns[5]].values[i + 2])
                    self.HR[i] = 0
                    self.load[i] = 0
                    self.VE[i] = float(df[df.columns[6]].values[i + 2])
                    self.Rf[i] = float(df[df.columns[1]].values[i + 2])
                    self.PetO2[i] = float(df[df.columns[16]].values[i + 2])
                    self.PetCO2[i] = float(df[df.columns[17]].values[i + 2])
                    i += 1
                except:
                    pass

        if self.metabolimeter == 'cortex_bruce_3':

            f = open(self.filename + self.data_extension, encoding="utf8", errors="ignore")
            df = read_csv(f, header=1)

            try:
                starting_index = df.loc[df['Unnamed: 0'] == 't'].index
                ref_string = 'Unnamed: 0'
            except:
                starting_index = df.loc[df['Rsultats de TCP'] == 't'].index
                ref_string = 'Rsultats de TCP'

            ventilatory_data = df[starting_index[0]:-1]
            # rename columns
            ventilatory_data.columns = ventilatory_data.iloc[0]

            # print('Weight')
            self.weight = float(df[df[ref_string] == 'Poids'].values[0, 2].split(' ')[0].replace(',', '.'))
            # print('Height')
            self.height = float(df[df[ref_string] == 'Taille'].values[0, 2].split(' ')[0])
            # print('Age')
            self.age = 2020 - float(df[df[ref_string] == 'Date de Naissance'].values[0, 2].split('/')[2])
            # print('Gender')
            if df[df[ref_string] == 'Sexe'].values[0, 2].split(' ')[0] == 'femme':
                self.gender = 'F'
            else:
                self.gender = 'M'

            # print('Cortex with VT data inside the same file -> no need to have labels')
            VT_index = df.loc[df[ref_string] == 'Variable'].index
            threshold_data = df[VT_index[0]:VT_index[0] + 21]

            use_WR = False
            use_HR = False

            try:
                WRVT1 = float(threshold_data[threshold_data[ref_string] == 'TT'].values[0][5])
                try:
                    WRVT2 = float(threshold_data[threshold_data[ref_string] == 'TT'].values[0][8])
                except:
                    WRVT2 = 1000
                use_WR = True
            except:
                HRVT1 = float(threshold_data[threshold_data[ref_string] == 'FC'].values[0][5])
                try:
                    HRVT2 = float(threshold_data[threshold_data[ref_string] == 'FC'].values[0][8])
                except:
                    HRVT2 = 1000
                use_HR = True

            # print('Reading data')
            n_rows = len(ventilatory_data.index)
            # initialise variables
            self.time = np.zeros((n_rows - 3,), dtype=np.float32)
            self.VO2 = np.zeros((n_rows - 3,), dtype=np.float32)
            self.VCO2 = np.zeros((n_rows - 3,), dtype=np.float32)
            self.HR = np.zeros((n_rows - 3,), dtype=np.float32)
            self.Rf = np.zeros((n_rows - 3,), dtype=np.float32)
            self.VE = np.zeros((n_rows - 3,), dtype=np.float32)
            self.PetO2 = np.zeros((n_rows - 3,), dtype=np.float32)
            self.PetCO2 = np.zeros((n_rows - 3,), dtype=np.float32)
            self.load = np.zeros((n_rows - 3,), dtype=np.float32)

            i = 0
            for time_sec in ventilatory_data.t.values[2:-1]:
                # print(i)
                try:
                    self.time[i] = utilities.get_sec(time_sec)
                    self.VO2[i] = float(ventilatory_data[ventilatory_data.columns[3]].values[i + 2]) * 1000
                    self.VCO2[i] = float(ventilatory_data[ventilatory_data.columns[78]].values[i + 2]) * 1000
                    self.HR[i] = float(ventilatory_data[ventilatory_data.columns[6]].values[i + 2])

                    try:
                        self.load[i] = float(ventilatory_data[ventilatory_data.columns[7]].values[i + 2])
                    except:
                        self.load[i] = 0

                    self.VE[i] = float(ventilatory_data[ventilatory_data.columns[11]].values[i + 2])
                    self.Rf[i] = float(ventilatory_data[ventilatory_data.columns[13]].values[i + 2])
                    self.PetO2[i] = float(ventilatory_data[ventilatory_data.columns[62]].values[i + 2])
                    self.PetCO2[i] = float(ventilatory_data[ventilatory_data.columns[61]].values[i + 2])
                    i += 1
                except:
                    pass

            # self.time = self.time - self.time[0]
            if use_WR:
                VT1_index = np.where(self.load == int(WRVT1))[0][0]
                try:
                    VT2_index = np.where(self.load == int(WRVT2))[0][0]
                except:
                    VT2_index = -1
            elif use_HR:
                VT1_index = np.where(self.HR == int(HRVT1))[0][0]
                try:
                    VT2_index = np.where(self.HR == int(HRVT2))[0][0]
                except:
                    VT2_index = -1

            self.VT1 = self.time[VT1_index]
            self.VT2 = self.time[VT2_index]

        if self.metabolimeter == 'cortex':

            df = self.df

            try:
                starting_index = df.loc[df['Unnamed: 0'] == 't'].index
                ref_string = 'Unnamed: 0'
            except:
                try:
                    starting_index = df.loc[df['CPET Results'] == 't'].index
                    ref_string = 'CPET Results'
                except:
                    starting_index = df.loc[df['Ergebnisse des Spiroergometrietests'] == 't'].index
                    ref_string = 'Ergebnisse des Spiroergometrietests'

            ventilatory_data = df[starting_index[0]:-1]
            # rename columns
            ventilatory_data.columns = ventilatory_data.iloc[0]

            # print('Weight')
            try:
                self.weight = float(df[df[ref_string] == 'Weight'].values[0, 2].split(' ')[0])
            except:
                self.height = float(df[df[ref_string] == 'Gewicht'].values[0, 2].split(' ')[0].split(',')[0])
            # print('Height')
            try:
                self.height = float(df[df[ref_string] == 'Height'].values[0, 2].split(' ')[0])
            except:
                self.height = float(df[df[ref_string] == 'Größe'].values[0, 2].split(' ')[0])
            # print('Age')
            try:
                self.age = 2020 - float(df[df[ref_string] == 'Date of Birth'].values[0, 2].split('/')[2])
            except:
                self.age = 2020 - float(df[df[ref_string] == 'Geburtsdatum'].values[0, 2].split('.')[2])
            # print('Gender')
            try:
                if df[df[ref_string] == 'Sex'].values[0, 2].split(' ')[0] == 'male':
                    self.gender = 'M'
                else:
                    self.gender = 'F'
            except:
                self.gender = 'M'

            # print('Cortex with VT data inside the same file -> no need to have labels')
            try:
                VT_index = df.loc[df[ref_string] == 'Variable'].index
                threshold_data = df[VT_index[0]:VT_index[0] + 21]

                use_WR = False
                use_HR = False

                try:
                    WRVT1 = float(threshold_data[threshold_data[ref_string] == 'WR'].values[0][5])
                    try:
                        WRVT2 = float(threshold_data[threshold_data[ref_string] == 'WR'].values[0][7])
                    except:
                        WRVT2 = 1000
                    use_WR = True
                except:
                    HRVT1 = float(threshold_data[threshold_data[ref_string] == 'HR'].values[0][5])
                    try:
                        HRVT2 = float(threshold_data[threshold_data[ref_string] == 'HR'].values[0][7])
                    except:
                        HRVT2 = 1000
                    use_HR = True
            except:
                pass

            # print('Reading data')
            n_rows = len(ventilatory_data.index)
            # initialise variables
            self.time = np.zeros((n_rows - 3,), dtype=np.float32)
            self.VO2 = np.zeros((n_rows - 3,), dtype=np.float32)
            self.VCO2 = np.zeros((n_rows - 3,), dtype=np.float32)
            self.HR = np.zeros((n_rows - 3,), dtype=np.float32)
            self.Rf = np.zeros((n_rows - 3,), dtype=np.float32)
            self.VE = np.zeros((n_rows - 3,), dtype=np.float32)
            self.PetO2 = np.zeros((n_rows - 3,), dtype=np.float32)
            self.PetCO2 = np.zeros((n_rows - 3,), dtype=np.float32)
            self.load = np.zeros((n_rows - 3,), dtype=np.float32)

            i = 0
            for time_sec in ventilatory_data.t.values[2:-1]:
                # print(i)
                try:
                    self.time[i] = utilities.get_sec(time_sec)
                except:
                    self.time[i] = np.nan

                self.VO2[i] = float(ventilatory_data[ventilatory_data.columns[3]].values[i+2]) * 1000
                self.VCO2[i] = float(ventilatory_data[ventilatory_data.columns[4]].values[i+2]) * 1000

                try:
                    self.HR[i] = float(ventilatory_data[ventilatory_data.columns[7]].values[i+2])
                except:
                    self.HR[i] = 0

                try:
                    self.load[i] = float(ventilatory_data[ventilatory_data.columns[8]].values[i+2])
                except:
                    self.load[i] = 0
                self.VE[i] = float(ventilatory_data[ventilatory_data.columns[11]].values[i+2])
                self.Rf[i] = float(ventilatory_data[ventilatory_data.columns[13]].values[i + 2])
                self.PetO2[i] = float(ventilatory_data[ventilatory_data.columns[14]].values[i + 2])
                self.PetCO2[i] = float(ventilatory_data[ventilatory_data.columns[15]].values[i + 2])

                i += 1

            VT1_index = 0
            VT2_index = 0

            try:
                # self.time = self.time - self.time[0]
                if use_WR:
                    VT1_index = np.where(self.load == int(WRVT1))[0][0]
                    try:
                        VT2_index = np.where(self.load == int(WRVT2))[0][0]
                    except:
                        VT2_index = -1
                elif use_HR:
                    VT1_index = np.where(self.HR == int(HRVT1))[0][0]
                    try:
                        VT2_index = np.where(self.HR == int(HRVT2))[0][0]
                    except:
                        VT2_index = -1
            except:
                pass

            self.VT1 = self.time[VT1_index]
            self.VT2 = self.time[VT2_index]

        if self.metabolimeter == 'unisbz':

            df = self.df

            self.age = float(0)
            # print('Gender')
            self.gender = 'N'
            # print('Weight')
            self.weight = float(0)
            # print('Height')
            self.height = float(0)

            # print('Reading data')
            n_rows = len(df.index)
            self.time = np.zeros((n_rows - 2,), dtype=np.float32)
            self.VO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.VCO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.HR = np.zeros((n_rows - 2,), dtype=np.float32)
            self.Rf = np.zeros((n_rows - 2,), dtype=np.float32)
            self.VE = np.zeros((n_rows - 2,), dtype=np.float32)
            self.PetO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.PetCO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.load = np.zeros((n_rows - 2,), dtype=np.float32)

            # get the time in seconds
            # convert time data from "HH:MM:SS" to seconds
            for i in df.index[0:n_rows-2]:
                try:
                    self.time[i] = utilities.get_sec(df[df.columns[0]].values[i])
                    self.VO2[i] = float(df[df.columns[14]].values[i])
                    self.VCO2[i] = float(df[df.columns[15]].values[i])
                    self.VE[i] = float(df[df.columns[11]].values[i])
                    self.HR[i] = float(df[df.columns[3]].values[i])
                    self.PetO2[i] = float(df[df.columns[8]].values[i])
                    self.PetCO2[i] = float(df[df.columns[7]].values[i])
                    self.Rf[i] = float(df[df.columns[6]].values[i])
                    self.load[i] = float(df[df.columns[1]].values[i])
                except:
                    self.time[i] = np.nan
                    self.VO2[i] = np.nan
                    self.VCO2[i] = np.nan
                    self.VE[i] = np.nan
                    self.HR[i] = np.nan
                    self.PetO2[i] = np.nan
                    self.PetCO2[i] = np.nan
                    self.Rf[i] = np.nan
                    self.load[i] = np.nan

        if self.metabolimeter == 'cosmed':

            df = self.df

            # print('Reading age')
            self.age = float(df.values[3, 1])
            # print('Gender')
            try:
                self.gender = df.values[2, 1][0]
            except:
                self.gender = 'M' # default to male
            # print('Weight')
            self.weight = float(df.values[5, 1])
            # print('Height')
            self.height = float(df.values[4, 1])

            # print('Reading data')
            n_rows = len(df.index)
            self.time = np.zeros((n_rows - 2,), dtype=np.float32)
            self.VO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.VCO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.HR = np.zeros((n_rows - 2,), dtype=np.float32)
            self.Rf = np.zeros((n_rows - 2,), dtype=np.float32)
            self.VE = np.zeros((n_rows - 2,), dtype=np.float32)
            self.PetO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.PetCO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.load = np.zeros((n_rows - 2,), dtype=np.float32)

            # get the time in seconds
            # convert time data from "HH:MM:SS" to seconds
            for i in df.index[2:n_rows]:
                try:
                    if df.t[0] == 's':
                        try:
                            self.time[i - 2] = int(round((float(df.t.values[i]))*86400))
                        except:
                            self.time[i - 2] = utilities.get_sec(df.t.values[i])
                    else:
                        self.time[i - 2] = utilities.get_sec(df.t.values[i])
                except:
                    self.time[i - 2] = np.nan
                self.VO2[i - 2] = float(df.VO2.values[i])
                self.VCO2[i - 2] = float(df.VCO2.values[i])
                try:
                    self.HR[i - 2] = float(df.HR.values[i])
                except:
                    self.HR[i - 2] = float(df.HF.values[i])
                try:
                    self.Rf[i - 2] = float(df.Rf.values[i])
                except:
                    try:
                        self.Rf[i - 2] = float(df['F.R'].values[i])
                    except:
                        self.Rf[i - 2] = float(df['Af'].values[i])
                try:
                    self.VE[i - 2] = float(df.VE.values[i])
                except:
                    self.VE[i - 2] = float(df.VE_ergo.values[i])
                self.PetO2[i - 2] = float(df.PetO2.values[i])
                self.PetCO2[i - 2] = float(df.PetCO2.values[i])

        if self.metabolimeter == 'cortex_bruce':

            df = self.df
            print('Reading data Bruce file!!!!')

            starting_index = df.loc[df[df.columns[0]] == 'Messzeit'].index+2
            n_rows = df[df.columns[0]].last_valid_index()-1

            # print('Reading age')
            my_array = np.where(df[df.columns[6]].notnull() == True, df.index, 0)
            res = next(x for x, val in enumerate(my_array) if val > 0.6)
            try:
                self.age = float(0)
                self.age = 2020 - int(df.iloc[4][2].split('.')[2])
                # print('Gender')
                self.gender = 'N'
                if str(df.iloc[3][2])[0] == 'm':
                    self.gender = 'M'
                if str(df.iloc[3][2])[0] == 'f':
                    self.gender = 'F'
                # print('Weight')
                self.weight = float(0)
                self.weight = float(df.iloc[6][2])
                # print('Height')
                self.height = float(0)
                self.height = float(int(df.iloc[5][2])/100)
            except:
                    self.age = float(0)
                    self.height = float(0)
                    self.gender = 'N'
                    self.weight = float(0)

            self.time = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.VO2 = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.VCO2 = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.HR = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.Rf = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.VE = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.PetO2 = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.PetCO2 = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.load = np.zeros((n_rows - starting_index[0],), dtype=np.float32)

            for i in np.arange(starting_index[0], n_rows - starting_index[0]):
                try:
                    self.time[i - starting_index[0]] = float(utilities.get_sec(df[df.columns[0]].values[i]))
                    self.VO2[i - starting_index[0]] = float(df[df.columns[10]].values[i]) * 1000
                    self.VCO2[i - starting_index[0]] = float(df[df.columns[11]].values[i]) * 1000
                    self.VE[i - starting_index[0]] = float(df[df.columns[4]].values[i])
                    self.HR[i - starting_index[0]] = float(df[df.columns[18]].values[i])
                    self.Rf[i - starting_index[0]] = float(df[df.columns[3]].values[i])
                    self.PetO2[i - starting_index[0]] = float(df[df.columns[8]].values[i])
                    self.PetCO2[i - starting_index[0]] = float(df[df.columns[9]].values[i])
                except:
                    self.time[i - starting_index[0]] = np.nan
                    self.VO2[i - starting_index[0]] = np.nan
                    self.VCO2[i - starting_index[0]] = np.nan
                    self.VE[i - starting_index[0]] = np.nan
                    self.HR[i - starting_index[0]] = np.nan
                    self.Rf[i - starting_index[0]] = np.nan
                    self.PetO2[i - starting_index[0]] = np.nan
                    self.PetCO2[i - starting_index[0]] = np.nan

        if self.metabolimeter == 'cortex_bruce_2':

            df = self.df
            print('Reading data Bruce file type 2!!!!')

            starting_index = df.loc[df[df.columns[0]] == 't'].index+2
            n_rows = df[df.columns[0]].last_valid_index()-1

            # print('Reading age')
            my_array = np.where(df[df.columns[6]].notnull() == True, df.index, 0)
            res = next(x for x, val in enumerate(my_array) if val > 0.6)
            try:
                self.age = float(0)
                self.age = 2020 - int(df.iloc[6][2].split('.')[2])
                # print('Gender')
                self.gender = 'N'
                if str(df.iloc[5][2])[0] == 'm':
                    self.gender = 'M'
                if str(df.iloc[5][2])[0] == 'f':
                    self.gender = 'F'
                # print('Weight')
                self.weight = float(0)
                self.weight = float(df.iloc[8][2])
                # print('Height')
                self.height = float(0)
                self.height = float(int(df.iloc[7][2])/100)
            except:
                    self.age = float(0)
                    self.height = float(0)
                    self.gender = 'N'
                    self.weight = float(0)

            self.time = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.VO2 = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.VCO2 = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.HR = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.Rf = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.VE = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.PetO2 = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.PetCO2 = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.load = np.zeros((n_rows - starting_index[0],), dtype=np.float32)

            for i in np.arange(starting_index[0], n_rows - starting_index[0]):
                try:
                    self.time[i - starting_index[0]] = float(utilities.get_sec(df[df.columns[0]].values[i]))
                    self.VO2[i - starting_index[0]] = float(df[df.columns[13]].values[i]) * 1000
                    self.VCO2[i - starting_index[0]] = float(df[df.columns[14]].values[i]) * 1000
                    self.VE[i - starting_index[0]] = float(df[df.columns[7]].values[i])
                    self.HR[i - starting_index[0]] = float(df[df.columns[21]].values[i])
                    self.Rf[i - starting_index[0]] = float(df[df.columns[3]].values[i])
                    self.PetO2[i - starting_index[0]] = float(df[df.columns[11]].values[i])
                    self.PetCO2[i - starting_index[0]] = float(df[df.columns[12]].values[i])
                except:
                    self.time[i - starting_index[0]] = np.nan
                    self.VO2[i - starting_index[0]] = np.nan
                    self.VCO2[i - starting_index[0]] = np.nan
                    self.VE[i - starting_index[0]] = np.nan
                    self.HR[i - starting_index[0]] = np.nan
                    self.Rf[i - starting_index[0]] = np.nan
                    self.PetO2[i - starting_index[0]] = np.nan
                    self.PetCO2[i - starting_index[0]] = np.nan

        if self.metabolimeter == 'mourot':

            df = self.df
            # print('Reading data Mourot file!!!!')

            starting_index = df.loc[df[df.columns[0]] == 'Temps'].index+2
            n_rows = df[df.columns[0]].last_valid_index()-1

            # print('Reading age')
            my_array = np.where(df[df.columns[6]].notnull() == True, df.index, 0)
            res = next(x for x, val in enumerate(my_array) if val > 0.6)
            try:
                self.age = float(0)
                self.age = float(df[df.columns[6]][res])
                # print('Gender')
                self.gender = 'N'
                if df[df.columns[18]][res][0] == 'M':
                    self.gender = 'M'
                if df[df.columns[18]][res][0] == 'F':
                    self.gender = 'F'
                # print('Weight')
                self.weight = float(0)
                self.weight = float(df[df.columns[27]][res])
                # print('Height')
                self.height = float(0)
                self.height = float(df[df.columns[22]][res])
            except:
                try:
                    res = res + 2
                    self.age = float(0)
                    self.age = float(df[df.columns[6]][res])
                    # print('Gender')
                    self.gender = 'N'
                    if df[df.columns[18]][res][0] == 'M':
                        self.gender = 'M'
                    if df[df.columns[18]][res][0] == 'F':
                        self.gender = 'F'
                    # print('Weight')
                    self.weight = float(0)
                    self.weight = float(df[df.columns[27]][res])
                    # print('Height')
                    self.height = float(0)
                    self.height = float(df[df.columns[22]][res])
                except:
                    self.age = float(0)
                    self.height = float(0)
                    self.gender = 'N'

            self.time = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.VO2 = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.VCO2 = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.HR = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.Rf = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.VE = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.PetO2 = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.PetCO2 = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.load = np.zeros((n_rows - starting_index[0],), dtype=np.float32)

            for i in np.arange(starting_index[0],n_rows - starting_index[0]):
                try:
                    self.time[i - starting_index[0]] = float(df[df.columns[0]].values[i]) * 60
                    self.VO2[i - starting_index[0]] = float(df[df.columns[9]].values[i]) * 1000
                    self.VCO2[i - starting_index[0]] = float(df[df.columns[11]].values[i]) * 1000
                    self.VE[i - starting_index[0]] = float(df[df.columns[22]].values[i])
                    self.HR[i - starting_index[0]] = float(df[df.columns[27]].values[i])
                    self.PetO2[i - starting_index[0]] = float(df[df.columns[20]].values[i]) * 760
                    self.PetCO2[i - starting_index[0]] = float(df[df.columns[21]].values[i]) * 760
                except:
                    self.time[i - starting_index[0]] = np.nan
                    self.VO2[i - starting_index[0]] = np.nan
                    self.VCO2[i - starting_index[0]] = np.nan
                    self.VE[i - starting_index[0]] = np.nan
                    self.HR[i - starting_index[0]] = np.nan
                    self.PetO2[i - starting_index[0]] = np.nan
                    self.PetCO2[i - starting_index[0]] = np.nan

        if self.metabolimeter == 'mourot_COPD':

            df = self.df
            print('Reading data Mourot file with no age nor BMI data !!!!')

            starting_index = [1]
            n_rows = df[df.columns[0]].last_valid_index() - 3

            # print('Reading age')
            my_array = np.where(df[df.columns[6]].notnull() == True, df.index, 0)
            res = next(x for x, val in enumerate(my_array) if val > 0.6)

            self.age = float(0)
            # print('Gender')
            self.gender = 'N'
            # print('Weight')
            self.weight = float(0)
            # print('Height')
            self.height = float(0)

            self.time = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.VO2 = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.VCO2 = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.HR = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.VE = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.PetO2 = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.PetCO2 = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.load = np.zeros((n_rows - starting_index[0],), dtype=np.float32)

            for i in np.arange(starting_index[0],n_rows):
                try:
                    self.time[i - starting_index[0]] = utilities.get_sec(df[df.columns[0]].values[i])
                except:
                    self.time[i - starting_index[0]] = np.nan
                try:
                    # print(i)
                    self.VO2[i - starting_index[0]] = float(df[df.columns[8]].values[i]) * 1000
                    if df[df.columns[10]].values[i] == ' ':
                        self.VCO2[i - starting_index[0]] = float(df[df.columns[4]].values[i])/float(df[df.columns[12]].values[i]) * 1000
                    else:
                        self.VCO2[i - starting_index[0]] = float(df[df.columns[10]].values[i]) * 1000
                    self.VE[i - starting_index[0]] = float(df[df.columns[4]].values[i])
                    self.HR[i - starting_index[0]] = float(df[df.columns[3]].values[i])
                    self.PetO2[i - starting_index[0]] = float(df[df.columns[15]].values[i]) * 7.60
                    self.PetCO2[i - starting_index[0]] = float(df[df.columns[17]].values[i])
                except:
                    self.time[i - starting_index[0]] = np.nan
                    self.VO2[i - starting_index[0]] = np.nan
                    self.VCO2[i - starting_index[0]] = np.nan
                    self.VE[i - starting_index[0]] = np.nan
                    self.HR[i - starting_index[0]] = np.nan
                    self.PetO2[i - starting_index[0]] = np.nan
                    self.PetCO2[i - starting_index[0]] = np.nan

            self.Rf = self.time * 0

        if self.metabolimeter == 'low':

            df = self.df

            # print('Reading age')
            self.age = float(30)
            # print('Gender')
            self.gender = 'N'
            # print('Weight')
            self.weight = float(50)
            # print('Height')
            self.height = float(180)

            # print('Reading data')
            n_rows = len(df.index)
            self.time = np.zeros((n_rows - 2,), dtype=np.float32)
            self.VO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.VCO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.HR = np.zeros((n_rows - 2,), dtype=np.float32)
            self.Rf = np.zeros((n_rows - 2,), dtype=np.float32)
            self.VE = np.zeros((n_rows - 2,), dtype=np.float32)
            self.PetO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.PetCO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.load = np.zeros((n_rows - 2,), dtype=np.float32)

            # get the time in seconds
            # convert time data from "HH:MM:SS" to seconds
            for i in df.index[2:n_rows]:
                try:
                    self.time[i - 2] = utilities.get_sec(df[df.columns[0]].values[i])
                    self.VO2[i - 2] = float(df[df.columns[4]].values[i])
                    self.VCO2[i - 2] = float(df[df.columns[5]].values[i])
                    self.VE[i - 2] = float(df[df.columns[3]].values[i])
                    self.HR[i - 2] = float(df[df.columns[2]].values[i])
                except:
                    self.time[i - 2] = np.nan
                    self.VO2[i - 2] = np.nan
                    self.VCO2[i - 2] = np.nan
                    self.VE[i - 2] = np.nan
                    self.HR[i - 2] = np.nan

        if self.metabolimeter == 'VO2Master':

            df = self.df

            # print('Reading age')
            self.age = float(30)
            # print('Gender')
            self.gender = 'N'
            # print('Weight')
            self.weight = float(50)
            # print('Height')
            self.height = float(180)

            # print('Reading data')
            n_rows = len(df.index)
            self.time = np.zeros((n_rows - 2,), dtype=np.float32)
            self.VO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.VCO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.HR = np.zeros((n_rows - 2,), dtype=np.float32)
            self.Rf = np.zeros((n_rows - 2,), dtype=np.float32)
            self.VE = np.zeros((n_rows - 2,), dtype=np.float32)
            self.PetO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.PetCO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.load = np.zeros((n_rows - 2,), dtype=np.float32)

            # get the time in seconds
            # convert time data from "HH:MM:SS" to seconds
            for i in df.index[1:n_rows]:
                try:
                    self.time[i - 2] = utilities.get_sec(df[df.columns[1]].values[i])
                    self.VO2[i - 2] = float(df[df.columns[3]].values[i])
                    self.VE[i - 2] = float(df[df.columns[6]].values[i])
                    self.Rf[i - 2] = float(df[df.columns[4]].values[i])
                    self.PetO2[i - 2] = float(df[df.columns[8]].values[i]) * 7.60
                except:
                    self.time[i - 2] = np.nan
                    self.VO2[i - 2] = np.nan
                    self.VCO2[i - 2] = np.nan
                    self.VE[i - 2] = np.nan
                    self.HR[i - 2] = np.nan

        if self.metabolimeter == 'centro-monzino':

            df = self.df

            # print('Reading age')
            self.age = float(30)
            # print('Gender')
            self.gender = 'N'
            # print('Weight')
            self.weight = float(50)
            # print('Height')
            self.height = float(180)

            # print('Reading data')
            n_rows = len(df.index)
            self.time = np.zeros((n_rows - 2,), dtype=np.float32)
            self.VO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.VCO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.HR = np.zeros((n_rows - 2,), dtype=np.float32)
            self.Rf = np.zeros((n_rows - 2,), dtype=np.float32)
            self.VE = np.zeros((n_rows - 2,), dtype=np.float32)
            self.PetO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.PetCO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.load = np.zeros((n_rows - 2,), dtype=np.float32)

            # get the time in seconds
            # convert time data from "HH:MM:SS" to seconds
            for i in df.index[2:n_rows]:
                try:
                    self.time[i - 2] = utilities.get_sec(df[df.columns[0]].values[i])
                    self.VO2[i - 2] = float(df[df.columns[2]].values[i])*1000
                    self.VCO2[i - 2] = float(df[df.columns[3]].values[i])*1000
                    self.VE[i - 2] = float(df[df.columns[4]].values[i])
                    self.HR[i - 2] = float(df[df.columns[9]].values[i])
                    self.PetO2[i - 2] = float(df[df.columns[5]].values[i])
                    self.PetCO2[i - 2] = float(df[df.columns[6]].values[i])
                except:
                    self.time[i - 2] = np.nan
                    self.VO2[i - 2] = np.nan
                    self.VCO2[i - 2] = np.nan
                    self.VE[i - 2] = np.nan
                    self.HR[i - 2] = np.nan

        if self.metabolimeter == 'mourot_cardiac':

            data = read_csv(self.filename + self.data_extension, encoding=result['encoding'], header=0, nrows=0, sep=',')
            d = gender.Detector()

            starting_index = [1]
            n_rows = df[df.columns[0]].last_valid_index() - 3

            # print('Reading age weight and height (neutral gender)')
            test_date = int((data.columns[7].split('/'))[2])
            if test_date < 1900:
                test_date = test_date + 2000

            self.age = test_date - int((data.columns[4].split('/'))[2])
            self.gender = 'N'
            try:
                if d.get_gender(data.columns[1].split(' ')[1].capitalize(), 'france') == 'male':
                    self.gender = 'M'
                if d.get_gender(data.columns[1].split(' ')[1].capitalize(), 'france') == 'female':
                    self.gender = 'F'
            except:
                pass
            self.weight = float(data.columns[10])
            self.height = float(data.columns[13])

            # print('Ventilatory variables')

            self.time = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.VO2 = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.VCO2 = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.HR = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.Rf = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.VE = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.PetO2 = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.PetCO2 = np.zeros((n_rows - starting_index[0],), dtype=np.float32)
            self.load = np.zeros((n_rows - starting_index[0],), dtype=np.float32)

            for i in np.arange(starting_index[0], n_rows):
                try:
                    self.time[i - starting_index[0]] = float(df[df.columns[0]].values[i]) * 60
                    self.VO2[i - starting_index[0]] = float(df[df.columns[4]].values[i]) * 1000
                    self.VCO2[i - starting_index[0]] = float(df[df.columns[6]].values[i]) * 1000
                    self.VE[i - starting_index[0]] = float(df[df.columns[3]].values[i])
                    self.HR[i - starting_index[0]] = float(df[df.columns[14]].values[i])
                    self.PetO2[i - starting_index[0]] = float(df[df.columns[20]].values[i])
                    self.PetCO2[i - starting_index[0]] = float(df[df.columns[21]].values[i])
                    self.Rf[i - starting_index[0]] = float(df[df.columns[1]].values[i])
                except:
                    self.time[i - starting_index[0]] = np.nan
                    self.VO2[i - starting_index[0]] = np.nan
                    self.VCO2[i - starting_index[0]] = np.nan
                    self.VE[i - starting_index[0]] = np.nan
                    self.HR[i - starting_index[0]] = np.nan
                    self.PetO2[i - starting_index[0]] = np.nan
                    self.PetCO2[i - starting_index[0]] = np.nan
                    self.Rf[i - starting_index[0]] = np.nan

        if self.metabolimeter == 'generated_pyoxynet':
            self.time = df.time
            self.VO2 = df.VO2_I
            self.VCO2 = df.VCO2_I
            self.HR = df.HR_I
            self.Rf = df.RF_I
            self.VE = df.VE_I
            self.PetO2 = df.PetO2_I
            self.PetCO2 = df.PetCO2_I
            self.age = df.Age.mean()
            if df.gender.mean() == 1:
                self.gender = 'F'
            else:
                self.gender = 'M'
            # print('Weight')
            self.weight = df.weight.mean()
            # print('Height')
            self.height = df.height.mean()

        if self.metabolimeter == 'exercise_threshold_app':
            self.time = df.t
            self.VO2 = df.VO2
            self.VCO2 = df.VCO2
            self.HR = df.HR
            self.Rf = df.RF
            self.VE = df.VE
            self.PetO2 = df.PetO2
            self.PetCO2 = df.PetCO2
            # default to 45
            self.age = 45
            # default to N
            self.gender = 'N'
            # print('Weight')
            # default to 70
            self.weight = 70
            # print('Height')
            # default to 170
            self.height = 170

        if self.metabolimeter == 'uni_of_essex':

            # print('Reading age')
            self.age = float(30)
            # print('Gender')
            self.gender = 'N'
            # print('Weight')
            self.weight = float(60)
            # print('Height')
            self.height = float(170)

            # print('Reading data')
            n_rows = len(df.index)
            self.time = np.zeros((n_rows - 2,), dtype=np.float32)
            self.VO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.VCO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.HR = np.zeros((n_rows - 2,), dtype=np.float32)
            self.Rf = np.zeros((n_rows - 2,), dtype=np.float32)
            self.VE = np.zeros((n_rows - 2,), dtype=np.float32)
            self.PetO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.PetCO2 = np.zeros((n_rows - 2,), dtype=np.float32)
            self.load = np.zeros((n_rows - 2,), dtype=np.float32)

            # get the time in seconds
            # convert time data from "HH:MM:SS" to seconds
            for i in df.index[2:n_rows]:
                try:
                    self.time[i - 2] = utilities.get_sec(df[df.columns[0]].values[i])
                    self.VO2[i - 2] = float(df[df.columns[1]].values[i])
                    self.VCO2[i - 2] = float(df[df.columns[2]].values[i])
                    self.VE[i - 2] = float(df[df.columns[3]].values[i])
                    self.PetO2[i - 2] = float(df[df.columns[4]].values[i]) * 7.50062
                    self.PetCO2[i - 2] = float(df[df.columns[5]].values[i]) * 7.50062
                    self.Rf[i - 2] = float(df[df.columns[9]].values[i])
                except:
                    self.time[i - 2] = np.nan
                    self.VO2[i - 2] = np.nan
                    self.VCO2[i - 2] = np.nan
                    self.VE[i - 2] = np.nan
                    self.HR[i - 2] = np.nan
                    self.PetO2[i - 2] = np.nan
                    self.PetCO2[i - 2] = np.nan
                    self.Rf[i - 2] = np.nan

        self.time = np.nan_to_num(self.time)
        self.VO2 = np.nan_to_num(self.VO2)
        self.VCO2 = np.nan_to_num(self.VCO2)
        self.HR = np.nan_to_num(self.HR)
        self.Rf = np.nan_to_num(self.Rf)
        self.VE = np.nan_to_num(self.VE)
        self.PetO2 = np.nan_to_num(self.PetO2)
        self.PetCO2 = np.nan_to_num(self.PetCO2)

        # rolling averages
        from scipy.ndimage import uniform_filter1d
        filter_size=20

        if self.metabolimeter != 'low':

            self.Rf_F = uniform_filter1d(self.Rf, size=filter_size)
            self.PetO2_F = uniform_filter1d(self.PetO2, size=filter_size)
            self.PetCO2_F = uniform_filter1d(self.PetCO2, size=filter_size)

        if self.metabolimeter == 'low':
            self.Rf_F = np.zeros((n_rows - 2,), dtype=np.float32)
            self.PetO2_F = np.zeros((n_rows - 2,), dtype=np.float32)
            self.PetCO2_F = np.zeros((n_rows - 2,), dtype=np.float32)

        self.VO2_F = uniform_filter1d(self.VO2, size=filter_size)
        self.VCO2_F = uniform_filter1d(self.VCO2, size=filter_size)
        self.HR_F = uniform_filter1d(self.HR, size=filter_size)
        self.VE_F = uniform_filter1d(self.VE, size=filter_size)

        # maximal VE value (cutting after)
        max_VE_idx = max([(v, i) for i, v in enumerate(self.VE_F)])[1]
        self.time_F = self.time[:max_VE_idx]
        self.VO2_F = self.VO2_F[:max_VE_idx]
        self.VCO2_F = self.VCO2_F[:max_VE_idx]
        self.HR_F = self.HR_F[:max_VE_idx]
        self.Rf_F = self.Rf_F[:max_VE_idx]
        self.VE_F = self.VE_F[:max_VE_idx]
        self.PetO2_F = self.PetO2_F[:max_VE_idx]
        self.PetCO2_F = self.PetCO2_F[:max_VE_idx]

        self.time_I = np.arange(int(self.time_F[0]), self.time_F[-1])
        self.VO2_I = np.interp(self.time_I, self.time_F, self.VO2_F)
        self.VCO2_I = np.interp(self.time_I, self.time_F, self.VCO2_F)
        self.HR_I = np.interp(self.time_I, self.time_F, self.HR_F)
        self.Rf_I = np.interp(self.time_I, self.time_F, self.Rf_F)
        self.VE_I = np.interp(self.time_I, self.time_F, self.VE_F)
        self.PetO2_I = np.interp(self.time_I, self.time_F, self.PetO2_F)
        self.PetCO2_I = np.interp(self.time_I, self.time_F, self.PetCO2_F)

    def print_details(self):
        if self.gender == "M":
            print("The name of the participant is: " + self.name + '.\n'
                                                                   "He is " + str(
                self.age) + ' years old.\nWeight: ' + str(self.weight) + ' kg\nHeight: ' + str(self.height) + ' cm')
        if self.gender == "F":
            print("The name of the participant is: " + self.name + '.\n'    "She is " + str(
                self.age) + ' years old.\nWeight: ' + str(self.weight) + ' kg\nHeight: ' + str(self.height) + ' cm')

    def create_data_frame(self):

        self.data_frame = pd.DataFrame()

        self.data_frame['time'] = self.time_I
        self.data_frame['VO2_I'] = self.VO2_I
        self.data_frame['VCO2_I'] = self.VCO2_I
        self.data_frame['VE_I'] = self.VE_I
        self.data_frame['HR_I'] = self.HR_I
        self.data_frame['RF_I'] = self.Rf_I
        self.data_frame['PetO2_I'] = self.PetO2_I
        self.data_frame['PetCO2_I'] = self.PetCO2_I

        # variables normalised on VO2
        self.data_frame['VEVO2_I'] = self.VE_I / self.VO2_I
        self.data_frame['VCO2VO2_I'] = self.VCO2_I / self.VO2_I
        self.data_frame['PetO2VO2_I'] = self.PetO2_I / self.VO2_I
        self.data_frame['PetCO2VO2_I'] = self.PetCO2_I / self.VO2_I

        if self.metabolimeter == 'VO2Master':
            # putting to zero
            self.data_frame['VEVCO2_I'] = self.VCO2_I
        else:
            self.data_frame['VEVCO2_I'] = self.VE_I / self.VCO2_I
        self.data_frame['age'] = self.age

        # height in cm
        if self.height > 100:
            self.height = self.height/100
        else:
            pass

        self.data_frame['height'] = np.ones(self.time_I.shape) * self.height
        self.data_frame['weight'] = np.ones(self.time_I.shape) * self.weight

        if self.gender == 'F':
            self.data_frame['gender'] = np.ones(self.time_I.shape)
        else:
            self.data_frame['gender'] = -np.ones(self.time_I.shape)

        try:
            self.data_frame['domain'] = self.domain[:len(self.time_I)]
        except:
            self.data_frame['domain'] = self.time_I

        # compute fitness and age group
        self.data_frame['age_group'] = np.ones(self.time_I.shape) * 2
        self.data_frame['fitness_group'] = np.ones(self.time_I.shape) * 2
        try:
            if self.age < 40 and self.gender == 'M' or self.gender == 'N':
                self.data_frame['age_group'] = np.ones(self.time_I.shape)
                if max(self.data_frame.VO2_I)/self.weight < 39.5:
                    self.data_frame['fitness_group'] = np.ones(self.time_I.shape) * 1
                if max(self.data_frame.VO2_I)/self.weight > 48.3:
                    self.data_frame['fitness_group'] = np.ones(self.time_I.shape) * 3
                if max(self.data_frame.VO2_I)/self.weight >= 39.5 and \
                        max(self.data_frame.VO2_I)/self.weight <= 48.3:
                    self.data_frame['fitness_group'] = np.ones(self.time_I.shape) * 2

            if self.age >= 40 and self.age < 60  and self.gender == 'M' or self.gender == 'N':
                self.data_frame['age_group'] = np.ones(self.time_I.shape) * 2
                if max(self.data_frame.VO2_I)/self.weight < 34.8:
                    self.data_frame['fitness_group'] = np.ones(self.time_I.shape) * 1
                if max(self.data_frame.VO2_I)/self.weight > 43.3:
                    self.data_frame['fitness_group'] = np.ones(self.time_I.shape) * 3
                if max(self.data_frame.VO2_I)/self.weight >= 34.8 and \
                        max(self.data_frame.VO2_I)/self.weight <= 43.3:
                    self.data_frame['fitness_group'] = np.ones(self.time_I.shape) * 2

            if self.age >= 60  and self.gender == 'M' or self.gender == 'N':
                self.data_frame['age_group'] = np.ones(self.time_I.shape) * 3
                if max(self.data_frame.VO2_I)/self.weight < 28.8:
                    self.data_frame['fitness_group'] = np.ones(self.time_I.shape) * 1
                if max(self.data_frame.VO2_I)/self.weight > 36.7:
                    self.data_frame['fitness_group'] = np.ones(self.time_I.shape) * 3
                if max(self.data_frame.VO2_I)/self.weight >= 28.8 and \
                        max(self.data_frame.VO2_I)/self.weight <= 36.7:
                    self.data_frame['fitness_group'] = np.ones(self.time_I.shape) * 2

            if self.age < 40 and self.gender == 'F' or self.gender == 'N':
                self.data_frame['age_group'] = np.ones(self.time_I.shape)
                if max(self.data_frame.VO2_I)/self.weight < 33.8:
                    self.data_frame['fitness_group'] = np.ones(self.time_I.shape) * 1
                if max(self.data_frame.VO2_I)/self.weight > 42.4:
                    self.data_frame['fitness_group'] = np.ones(self.time_I.shape) * 3
                if max(self.data_frame.VO2_I)/self.weight >= 33.8 and \
                        max(self.data_frame.VO2_I)/self.weight <= 42.4:
                    self.data_frame['fitness_group'] = np.ones(self.time_I.shape) * 2

            if self.age >= 40 and self.age < 60  and self.gender == 'F' or self.gender == 'N':
                self.data_frame['age_group'] = np.ones(self.time_I.shape) * 2
                if max(self.data_frame.VO2_I)/self.weight < 32.3:
                    self.data_frame['fitness_group'] = np.ones(self.time_I.shape) * 1
                if max(self.data_frame.VO2_I)/self.weight > 39.6:
                    self.data_frame['fitness_group'] = np.ones(self.time_I.shape) * 3
                if max(self.data_frame.VO2_I)/self.weight >= 32.3 and \
                        max(self.data_frame.VO2_I)/self.weight <= 39.6:
                    self.data_frame['fitness_group'] = np.ones(self.time_I.shape) * 2

            if self.age >= 60  and self.gender == 'F' or self.gender == 'N':
                self.data_frame['age_group'] = np.ones(self.time_I.shape) * 3
                if max(self.data_frame.VO2_I)/self.weight < 25.3:
                    self.data_frame['fitness_group'] = np.ones(self.time_I.shape) * 1
                if max(self.data_frame.VO2_I)/self.weight > 30.6:
                    self.data_frame['fitness_group'] = np.ones(self.time_I.shape) * 3
                if max(self.data_frame.VO2_I)/self.weight >= 25.3 and \
                        max(self.data_frame.VO2_I)/self.weight <= 30.6:
                    self.data_frame['fitness_group'] = np.ones(self.time_I.shape) * 2
        except:
            self.data_frame['age_group'] = np.ones(self.time_I.shape) * 2
            self.data_frame['fitness_group'] = np.ones(self.time_I.shape) * 2

    def generate_pickle(self):
        self.data_frame.to_pickle('training_data/' + self.filename_cleared + '.pickle')
        print('Generating pickle file for ' + self.filename)

    def generate_csv(self):
        self.data_frame.to_csv('training_data/' + self.filename_cleared + '.csv')
        print('Generating csv file for ' + self.filename)

    def set_labels(self, label_data_frame):
        import math

        self.domain = np.empty([len(self.time_I),], int)

        if self.metabolimeter == 'cortex':

            self.domain[self.time_I < self.VT1] = -1 # low
            self.domain[(self.time_I >= self.VT1) & (self.time_I < self.VT2)] = 0 # moderate
            self.domain[self.time_I >= self.VT2] = 1 # high

        if self.metabolimeter == 'mourot' or self.metabolimeter == 'mourot_COPD' or self.metabolimeter == 'low' \
                or self.metabolimeter == 'cosmed':

            VT1 = int(label_data_frame[label_data_frame.id==self.filename_cleared].Feature1.values[0])

            if not math.isnan(VT1):
                try:
                    VT2 = int(label_data_frame[label_data_frame.id==self.filename_cleared].Feature2.values[0])
                except:
                    VT2 = self.time_I[-1]

            self.domain[self.time_I < VT1] = -1 # low
            self.domain[(self.time_I >= VT1) & (self.time_I < VT2)] = 0 # moderate
            self.domain[self.time_I >= VT2] = 1 # high

        if self.metabolimeter == 'mourot_cardiac':

            print(self.filename.split('/')[-1])

            VT1 = int(label_data_frame[label_data_frame.id==self.filename.split('/')[-1]].Feature1.values[0])

            # print(VT1)

            if not math.isnan(VT1):
                try:
                    VT2 = int(label_data_frame[label_data_frame.id==self.filename.split('/')[-1]].Feature2.values[0])
                except:
                    VT2 = self.time_I[-1]

            self.domain[self.time_I < VT1] = -1 # low
            self.domain[(self.time_I >= VT1) & (self.time_I < VT2)] = 0 # moderate
            self.domain[self.time_I >= VT2] = 1 # high

        if self.metabolimeter == 'unisbz':

            VT1 = int(label_data_frame[label_data_frame.id==int(self.filename_cleared)].Feature1.values[0])

            if not math.isnan(VT1):
                try:
                    VT2 = int(label_data_frame[label_data_frame.id==int(self.filename_cleared)].Feature2.values[0])
                except:
                    VT2 = self.time_I[-1]

            self.domain[self.time_I < VT1] = -1 # low
            self.domain[(self.time_I >= VT1) & (self.time_I < VT2)] = 0 # moderate
            self.domain[self.time_I >= VT2] = 1 # high


import numpy as np
from scipy.stats import beta

def PrintHello(hello='hello'):
    """This function prints to screen.

    Args:
       name (str):  The name to use.

    Returns:
       none

    """
    
    print(hello)
    
    return

def get_sec(time_str):
    """This function converts strings to time.

    Args:
       time_str (str):  The time str.

    Returns:
       total seconds (int)

    """

    my_time_str = time_str.replace('.', ':')
    try:
        h, m, s = my_time_str.split(':')
        return int(h) * 3600 + int(m) * 60 + int(s)
    except:
        try:
            h, m, s, ms = my_time_str.split(':')
            return int(h) * 3600 + int(m) * 60 + int(s)
        except:
            m, s = my_time_str.split(':')
            return int(m) * 60 + int(s)

def normalize(df, min_target=-1, max_target=1):
    """Pandas df normalisation

    Parameters:
        df (pd df) : input df

    Returns:
        result (pd df) : output df

    """

    result = df.copy()
    for feature_name in df.columns:
        max_value = df[feature_name].max()
        min_value = df[feature_name].min()
        # TODO: pelase check here the normalisation between 0 and 1 or -1 and 1
        # result[feature_name] = 2 * (df[feature_name] - min_value) / (max_value - min_value) - 1
        result[feature_name] = (df[feature_name] - min_value) / (max_value - min_value)
    result = result.fillna(0)

    return result

def load_explainer(tf_model):
    """Loads a deep explainer from the shap library and automatically loads the data to train it

    Parameters:
        tf_model (Model class) : deep learning model that deep explainer needs to explain

    Returns:
        explainer (deep) : deep explainer SHAP

    """

    import shap
    from pyoxynet import data_test
    import importlib_resources
    import pickle
    from keras.layers import Input
    from keras.models import Model as mdl
    from .model import Model

    train_data_X_binaries = importlib_resources.read_binary(data_test, 'train_data_X.pickle')
    open('/tmp/train_data_X.pickle', 'wb').write(train_data_X_binaries)
    with open('/tmp/train_data_X.pickle', 'rb') as f:
        training_data = pickle.load(f)

    shap.explainers._deep.deep_tf.op_handlers['AddV2'] = shap.explainers._deep.deep_tf.passthrough
    model = Model(n_classes=3, n_input=6)
    newInput = Input(batch_shape=(1, 40, 6))
    newOutputs = model(newInput)
    newModel = mdl(newInput, newOutputs)
    newModel.set_weights(tf_model.get_weights())
    # explainer = shap.DeepExplainer(newModel, training_data)
    # Gradient explainer faster and perhaps more robust when it comes to TF2 applications
    explainer = shap.GradientExplainer(newModel, training_data)

    return explainer

def compute_shap(explainer, df, n_inputs=6, past_points=40, shap_stride=20):
    """Computes SHAP values

    Parameters:
        explainer (explainer SHAP) : deep explainer that explains the deep learning model
        data (pd df) : data for the local explanation

    Returns:
        shap (dict) : SHAP values

    """

    import shap
    import pandas as pd
    import numpy as np

    # some adjustments to input df
    # TODO: create dedicated function for this (duplicated taks)
    df = df.drop_duplicates('time')
    df['timestamp'] = pd.to_datetime(df['time'], unit='s')
    df = df.set_index('timestamp')
    df = df.resample('1S').mean()
    df = df.interpolate()
    df['VO2_20s'] = df.VO2_I.rolling(20, win_type='triang', center=True).mean().fillna(method='bfill').fillna(
        method='ffill')
    df = df.reset_index()
    df = df.drop('timestamp', axis=1)

    if n_inputs == 6 and past_points == 40:
        if 'VCO2VO2_I' not in df.columns:
            df['VCO2VO2_I'] = df['VCO2_I'].values/df['VO2_I'].values
        filter_vars = ['VCO2VO2_I', 'VE_I', 'PetO2_I', 'PetCO2_I', 'VEVO2_I', 'VEVCO2_I']
        X = df[filter_vars]
        XN = normalize(X)
        XN = XN.filter(filter_vars, axis=1)
    if n_inputs == 5 and past_points == 40:
        filter_vars = ['VO2_I', 'VE_I', 'PetO2_I', 'RF_I', 'VEVO2_I']
        X = df[filter_vars]
        XN = normalize(X)
        XN = XN.filter(filter_vars, axis=1)

    data_X = np.zeros([0, past_points, n_inputs])
    for i in np.arange(1, XN.shape[0] - past_points, shap_stride):
        tmp_x = XN[i: i + past_points]
        tmp_x = np.expand_dims(tmp_x, 0)
        data_X = np.vstack((data_X, tmp_x))

    print('Computing SHAP values in progress')
    shap_values = explainer.shap_values(data_X)
    print('Done computing SHAP values!')

    tmp_hv = pd.DataFrame(np.mean(shap_values[0][:, :, :], axis=1), columns=['hv_shap_v' + str(i) for i in np.arange(0, 6)])
    tmp_sv = pd.DataFrame(np.mean(shap_values[1][:, :, :], axis=1),
                          columns=['sv_shap_v' + str(i) for i in np.arange(0, 6)])
    tmp_md = pd.DataFrame(np.mean(shap_values[2][:, :, :], axis=1),
                          columns=['md_shap_v' + str(i) for i in np.arange(0, 6)])

    df_tmp = pd.concat([tmp_md, tmp_hv, tmp_sv], axis=1)

    return shap_values, df_tmp

def optimal_filter(t, y, my_lambda):
    """A bad ass optimisation filter
    
    Parameters: 
        t : array
            Independent coord array
        y : array
            Dependent coord array
        my_lambda : float
            Smoothing factor

    Returns:
        x : array
            Filtered variable
    
    """

    import numpy as np

    # be robust for non-monotonic x variables (made for the time in CPET specifically)
    for i in np.arange(1, len(t)-1):
        if t[i+1] == t[i]:
            t[i+1] = t[i] + 0.1

    h  = 0.5 * np.concatenate([[t[1]-t[0]], t[2:] - t[0:-2], [t[-1] - t[-2]]])
    # Robustness
    m = np.median(h[h > 0])
    # Assign the median to the zero elements
    h[h == 0] = m

    dg = np.divide(my_lambda,h)
    # symmetric tri-diagonal system
    a = - dg[1:]
    b = np.diff(t) + dg[0:-1] + dg[1:]
    u = np.diff(y)
    # Solution of the system
    n = len(u)

    for j in np.arange(0, n - 1):
        mu = a[j] / b[j]
        b[j + 1] = b[j + 1] - mu * a[j]
        u[j + 1] = u[j + 1] - mu * u[j]

    u[n-1] = u[n-1] / b[n-1]

    for j in np.arange(n-2, -1, -1):
        u[j] = (u[j] - a[j] * u[j + 1]) / b[j]

    # Retrieving solution
    x = np.empty([len(y), ])
    x[0] = y[0] + my_lambda * u[0] / h[0]

    for i in np.arange(0, n):
        x[i + 1] = x[i] + (t[i + 1] - t[i]) * u[i]

    return x

def load_tf_model(n_inputs=6, past_points=40, model='CNN'):
    """This function loads the saved tflite models.

    Args:
       n_inputs (int):  Number of input variables.
       past_points (int):  Number of past inputs in the time-series.

    Returns:
       interpreter (tflite interpreter) : handle on the TFLite interpreter

    """

    import importlib_resources
    import pickle
    from io import BytesIO
    from pyoxynet import regressor, TCN
    import tensorflow as tf
    import os

    if n_inputs == 6:
        if model == 'CNN':
            # load the classic Oxynet model configuration
            print('Classic Oxynet configuration model uploaded')
            saved_model_binaries = importlib_resources.read_binary(regressor, 'saved_model.pb')
            keras_metadata_model_binaries = importlib_resources.read_binary(regressor, 'keras_metadata.pb')
            variables_data_binaries = importlib_resources.read_binary(regressor, 'variables.data-00000-of-00001')
            variables_index_binaries = importlib_resources.read_binary(regressor, 'variables.index')
        if model == 'TCN':
            # load the classic Oxynet model configuration
            print('Classic Oxynet configuration model uploaded')
            saved_model_binaries = importlib_resources.read_binary(TCN, 'saved_model.pb')
            keras_metadata_model_binaries = importlib_resources.read_binary(TCN, 'keras_metadata.pb')
            variables_data_binaries = importlib_resources.read_binary(TCN, 'variables.data-00000-of-00001')
            variables_index_binaries = importlib_resources.read_binary(TCN, 'variables.index')
        if model == 'transformer':
            # load the classic Oxynet model configuration
            print('Classic Oxynet configuration model uploaded')
            tfl_model_binaries = importlib_resources.read_binary(regressor, 'transformer.pickle')
    if n_inputs == 5 and past_points == 40:
        # load the 5 input model configuration (e.g. in this case when on CO2 info is included)
        print('Specific configuration model uploaded (no VCO2 available)')
        tfl_model_binaries = importlib_resources.read_binary(regressor, 'tfl_model_5_40.pickle')

    try:
        if not os.path.isdir('/tmp/variables'):
            os.mkdir('/tmp/variables')
        open('/tmp/saved_model.pb', 'wb').write(saved_model_binaries)
        open('/tmp/keras_metadata.pb', 'wb').write(keras_metadata_model_binaries)
        open('/tmp/variables/variables.data-00000-of-00001', 'wb').write(variables_data_binaries)
        open('/tmp/variables/variables.index', 'wb').write(variables_index_binaries)

        from .model import Model, TCN
        if model == 'CNN':
            model = tf.keras.models.load_model('/tmp/')
            my_model = Model(n_classes=3, n_input=6)
            my_model.build(input_shape=(1, past_points, 6))
            my_model.set_weights(model.get_weights())
        if model == 'TCN':
            model = tf.keras.models.load_model('/tmp/')
            num_classes = 3  # Replace with the number of classes in your task
            num_filters = 64
            kernel_size = 3
            dilation_rates = [2 ** i for i in range(5)]
            my_model = TCN(num_classes, num_filters, kernel_size, dilation_rates)
            my_model.build(input_shape=(1, past_points, 6))
            my_model.set_weights(model.get_weights())

        return my_model
    except:
        print('Could not find a model that could satisfy the input size required')
        return None  

def load_tf_generator():
    """This function loads the saved tflite generator model.

    Args:
       None

    Returns:
       generator (tflite generator) : handle on the TFLite generator

    """

    import importlib_resources
    import pickle
    from io import BytesIO
    from pyoxynet import generator
    import tensorflow as tf
    import os

    # load the classic Oxynet model configuration
    print('Classic Oxynet configuration model uploaded')
    saved_model_binaries = importlib_resources.read_binary(generator, 'saved_model.pb')
    keras_metadata_model_binaries = importlib_resources.read_binary(generator, 'keras_metadata.pb')
    variables_data_binaries = importlib_resources.read_binary(generator, 'variables.data-00000-of-00001')
    variables_index_binaries = importlib_resources.read_binary(generator, 'variables.index')

    try:
        if not os.path.isdir('/tmp/variables'):
            os.mkdir('/tmp/variables')
        open('/tmp/saved_model.pb', 'wb').write(saved_model_binaries)
        open('/tmp/keras_metadata.pb', 'wb').write(keras_metadata_model_binaries)
        open('/tmp/variables/variables.data-00000-of-00001', 'wb').write(variables_data_binaries)
        open('/tmp/variables/variables.index', 'wb').write(variables_index_binaries)
        model = tf.keras.models.load_model('/tmp/')
        from .model import Generator
        my_model = Generator()
        # TODO: this is hardcoded
        my_model.build(input_shape=(1, 53))
        my_model.set_weights(model.get_weights())

        return my_model
    except:
        print('Could not find a model that could satisfy the input size required')
        return None

def pip_install_tflite():
    """Makes sure TFLite is installed by executing a pip install command from the command line (sub-optimal solution)

    Parameters: 
        none

    Returns:
        none

    """
    import os
    import pkg_resources
    installed_packages = pkg_resources.working_set
    installed_packages_list = sorted(["%s" % (i.key) for i in installed_packages])

    if 'tflite-runtime' in installed_packages_list:
        print('Tflite runtime already present in the package list (skipping)')
    else:
        os.system("pip install --extra-index-url https://google-coral.github.io/py-repo/ tflite_runtime")

def test_tfl_model(interpreter):
    """Test if the model is running correclty

    Parameters: 
        interpreter (loaded tf.lite.Interpreter) : Loaded interpreter TFLite model

    Returns:
        x (array) : Model output example

    """
    import numpy as np

    # Allocate tensors.
    interpreter.allocate_tensors()

    # Get input and output tensors.
    input_details = interpreter.get_input_details()
    output_details = interpreter.get_output_details()

    # Test the model on random input data.
    input_shape = input_details[0]['shape']
    input_data = np.array(np.random.random_sample(input_shape), dtype=np.float32)

    interpreter.set_tensor(input_details[0]['index'], input_data)
    interpreter.invoke()

    return interpreter.get_tensor(output_details[0]['index'])

def load_csv_data(csv_file='data_test.csv'):
    """Loads data from csv file (returns test data if no arguments)

    Parameters:
        n_inputs (int) :  Number of input variables.
        past_points (int) :  Number of past inputs in the time-series.

    Returns:
        df (pandas df) : Model output example

    """

    from importlib import resources
    import pandas as pd
    import pyoxynet.data_test

    if csv_file=='data_test.csv':
        import pkgutil
        from io import StringIO
        bytes_data = pkgutil.get_data('pyoxynet.data_test', csv_file)
        s = str(bytes_data, 'utf-8')
        data = StringIO(s)
        df = pd.read_csv(data)

    return df

def draw_real_test(resting='random'):
    """Draw a single data file from the directory containing all the real files

    Parameters:
        none

    Returns:
        df (pandas df) : Real data output

    """

    from importlib import resources
    import pandas as pd
    import pyoxynet.data_test
    import pkgutil
    from io import StringIO
    import random
    import numpy as np
    from datetime import datetime

    if resting == 'random':
        resting = random.randint(0, 1)

    if resting:
        file_index = str(random.randrange(1, 50))
        file_name = 'resting_real_test_' + file_index + '.csv'
        bytes_data = pkgutil.get_data('pyoxynet.data_test', file_name)
        print(file_name, ' from resting')
    else:
        file_index = str(random.randrange(1, 88))
        file_name = 'ramp_real_test_' + file_index + '.csv'
        bytes_data = pkgutil.get_data('pyoxynet.data_test', file_name)
        print(file_name, ' from ramp')

    s = str(bytes_data, 'utf-8')
    data = StringIO(s)
    df = pd.read_csv(data)

    if int(np.mean(df.fitness_group.values)) == 1:
        fitness_group = 'LOW'
    if int(np.mean(df.fitness_group.values)) == 2:
        fitness_group = 'MEDIUM'
    if int(np.mean(df.fitness_group.values)) == 3:
        fitness_group = 'HIGH'
    if int(np.mean(df.gender.values)) == -1:
        gender = 'MALE'
    if int(np.mean(df.gender.values)) == 1:
        gender = 'FEMALE'

    VT1 = df[np.diff(df.domain, prepend=-1) == 1].time.iloc[0]
    VT2 = df[np.diff(df.domain, prepend=-1) == 1].time.iloc[1]

    duration = len(df)

    df['VO2_I'] = df['VO2_I'] + [random.uniform(-100, 100) for i in np.arange(duration)]
    df['VCO2_I'] = df['VCO2_I'] + [random.uniform(-100, 100) for i in np.arange(duration)]
    df['VE_I'] = df['VE_I'] + [random.uniform(-2, 2) for i in np.arange(duration)]
    df['HR_I'] = df['HR_I'] + [random.uniform(-1, 1) for i in np.arange(duration)]
    df['RF_I'] = df['RF_I'] + [random.uniform(-2, 2) for i in np.arange(duration)]
    df['PetO2_I'] = df['PetO2_I'] + [random.uniform(-1, 1) for i in np.arange(duration)]
    df['PetCO2_I'] = df['PetCO2_I'] + [random.uniform(-1, 1) for i in np.arange(duration)]

    df['VEVO2_I'] = df['VE_I']/df['VO2_I']
    df['VEVCO2_I'] = df['VE_I']/df['VCO2_I']

    # Collect VO2 value at VT1 and VT2
    VO2VT1 = df.iloc[(df[df['domain'].diff().fillna(0) == 1].index[0] - 10):(df[df['domain'].diff().fillna(0) == 1].index[0] + 10)]['VO2_I'].mean()
    VO2VT2 = df.iloc[(df[df['domain'].diff().fillna(0) == 1].index[1] - 10):(
            df[df['domain'].diff().fillna(0) == 1].index[1] + 10)]['VO2_I'].mean()

    VO2_peak = df.VO2_I.rolling(20).mean().max()

    print('Data loaded for a ', gender, ' individual with ', fitness_group, ' fitness capacity.')
    print('Weight: ', int(np.mean(df.weight.values)), ' kg')
    print('Height: ', np.mean(df.height.values[0]), 'm')
    print('Age: ', int(np.mean(df.age.values)), 'y')
    print('VT1: ', str(VT1))
    print('VT2: ', str(VT2))
    print('VO2VT2: ', str(int(VO2VT1)))
    print('VO2VT2: ', str(int(VO2VT2)))

    data = dict()
    data['Age'] = str(int(df.age.values[0]))
    data['Height'] = str(df.height.values[0])
    data['Weight'] = str(int(df.weight.values[0]))
    data['Gender'] = gender
    data['Aerobic_fitness_level'] = fitness_group
    data['VT1'] = str(VT1)
    data['VT2'] = str(VT2)
    data['VO2VT1'] = str(int(VO2VT1))
    data['VO2VT2'] = str(int(VO2VT2))
    data['VO2max'] = str(int(VO2_peak))
    data['LT'] = str(int(VO2VT1))
    data['LT_vo2max'] = str(int((VO2VT1/VO2_peak)*100)) + '%'
    data['RCP'] = str(int(VO2VT2))
    data['RCP_vo2max'] = str(int((VO2VT2/VO2_peak) * 100)) + '%'
    data['id'] = 'real_#'
    data['noise_factor'] = 'NA'
    data['created'] = datetime.today().strftime("%m/%d/%Y - %H:%M:%S")
    data['resting'] = str(resting)

    df['breaths'] = (1/(df.RF_I/60))
    # df['recorded_timestamp'] = pd.to_datetime(df.time, unit='s')

    time_cum_sum = 0
    df_breath = pd.DataFrame()
    n = 0
    while time_cum_sum < (duration-20):
        try:
            tmp = df[(df.time >= time_cum_sum) & (df.time < (time_cum_sum + df.breaths.iloc[n]))].mean()
            df_breath = pd.concat([df_breath, pd.DataFrame(data=np.reshape(tmp.values, [1, len(df.columns)]),
                                                           columns=tmp.index.to_list())])
            time_cum_sum = (time_cum_sum + df.breaths.iloc[n])
            n = n + len(df[(df.time >= time_cum_sum) & (df.time < (time_cum_sum + df.breaths.iloc[n]))])
        except:
            break

    df_breath = df_breath.dropna()
    df_breath['time'] = df_breath['time'].astype(int)
    df_breath = df_breath.drop_duplicates('time')

    exercise_threshold_names = {"time": "t",
                                "VO2_I": "VO2",
                                "VCO2_I": "VCO2",
                                "VE_I": "VE",
                                "VCO2VO2_I": "R",
                                "VEVO2_I": "VE/VO2",
                                "VEVCO2_I": "VE/VCO2",
                                "PetO2_I": "PetO2",
                                "PetCO2_I": "PetCO2",
                                "HR_I": "HR",
                                "RF_I": "RF"}

    df_breath = df_breath.rename(columns=exercise_threshold_names)
    # create the dict for the exercise threshold app
    data['data'] = dict()
    data['data'] = df_breath.to_dict(orient='records')

    return df, data

def load_exercise_threshold_app_data(data_dict={}):
    """Loads data from data dict with format provided by https://www.exercisethresholds.com/

    Parameters:
        data_dict (dict) : Dictionary with format like test/exercise_threshold_app_test.json

    Returns:
        df (pandas df) : Pandas data frame with format that can be used by Pyoxynet for inference (columns needed: 'VO2_I', 'VCO2_I', 'VE_I', 'PetO2_I', 'PetCO2_I', 'VEVO2_I', 'VEVCO2_I')

    """

    import json
    import pandas as pd
    import numpy as np

    time = []
    VO2_I = []
    VCO2_I = []
    VCO2VO2_I = []
    VE_I = []
    PetO2_I = []
    PetCO2_I = []
    VEVO2_I = []
    VEVCO2_I = []
    HR_I = []

    for data_points_ in data_dict[0]['data']:
        time.append(data_points_['t'])
        VO2_I.append(data_points_['VO2'])
        VCO2_I.append(data_points_['VCO2'])
        VE_I.append(data_points_['VE'])
        PetO2_I.append(data_points_['PetO2'])
        PetCO2_I.append(data_points_['PetCO2'])
        VEVO2_I.append(data_points_['VE/VO2'])
        VEVCO2_I.append(data_points_['VE/VCO2'])
        VCO2VO2_I.append(data_points_['VCO2']/data_points_['VO2'])

    # transform in array
    time = np.asarray(time)
    VO2_I = np.asarray(VO2_I)
    VCO2_I = np.asarray(VCO2_I)
    VE_I = np.asarray(VE_I)
    PetO2_I = np.asarray(PetO2_I)
    PetCO2_I = np.asarray(PetCO2_I)
    VEVO2_I = np.asarray(VEVO2_I)
    VEVCO2_I = np.asarray(VEVCO2_I)
    VCO2VO2_I = np.asarray(VCO2VO2_I)

    # filter and interpolate
    time_I = np.arange(int(time[0]), int(time[-1]))
    VO2_I = np.interp(time_I, time, VO2_I)
    VCO2_I = np.interp(time_I, time, VCO2_I)
    VE_I = np.interp(time_I, time, VE_I)
    PetO2_I = np.interp(time_I, time, PetO2_I)
    PetCO2_I = np.interp(time_I, time, PetCO2_I)
    VEVO2_I = np.interp(time_I, time, VEVO2_I)
    VEVCO2_I = np.interp(time_I, time, VEVCO2_I)
    VCO2VO2_I = np.interp(time_I, time, VCO2VO2_I)

    # rolling averages to filter the data
    from scipy.ndimage import uniform_filter1d
    # This filter size should be equal to the one used by the Test class
    filter_size = 20

    VO2_I = uniform_filter1d(VO2_I, size=filter_size)
    VCO2_I = uniform_filter1d(VCO2_I, size=filter_size)
    VE_I = uniform_filter1d(VE_I, size=filter_size)
    PetO2_I = uniform_filter1d(PetO2_I, size=filter_size)
    PetCO2_I = uniform_filter1d(PetCO2_I, size=filter_size)
    VEVO2_I = uniform_filter1d(VEVO2_I, size=filter_size)
    VEVCO2_I = uniform_filter1d(VEVCO2_I, size=filter_size)
    VCO2VO2_I = uniform_filter1d(VCO2VO2_I, size=filter_size)

    df = pd.DataFrame()
    df['time'] = time_I
    df['VO2_I'] = VO2_I
    df['VCO2_I'] = VCO2_I
    df['VE_I'] = VE_I
    df['PetO2_I'] = PetO2_I
    df['PetCO2_I'] = PetCO2_I
    df['VEVO2_I'] = VEVO2_I
    df['VEVCO2_I'] = VEVCO2_I
    df['VCO2VO2_I'] = VCO2VO2_I
    df['HR_I'] = 0

    return df

def test_pyoxynet(input_df=[], n_inputs=6, past_points=40, model='CNN', plot=True):
    """Runs the pyoxynet inference

    Parameters: 
        n_inputs (int) : Number of inputs (deafult to Oxynet configuration)
        past_points (int) : Number of past points in the time series (deafult to Oxynet configuration)

    Returns:
        x (array) : Model output example

    """

    import numpy as np
    from uniplot import plot
    import pandas as pd
    from scipy import stats
    import json

    # import the data for the normalisation
    import pkgutil
    from io import StringIO
    bytes_data = pkgutil.get_data('pyoxynet.data_test', 'database_statistics_resting.csv')

    s = str(bytes_data, 'utf-8')
    data = StringIO(s)
    db_df = pd.read_csv(data)

    tf_model = load_tf_model(n_inputs=n_inputs, past_points=past_points, model=model)

    if len(input_df) == 0:
        print('Using default pyoxynet data')
        df = load_csv_data()
    else:
        df = input_df

    # some adjustments to input df
    # TODO: create dedicated function for this
    df = df.drop_duplicates('time')
    df['timestamp'] = pd.to_datetime(df['time'], unit='s')
    df = df.set_index('timestamp')
    df = df.resample('1S').mean()
    df = df.interpolate()
    df['VO2_20s'] = df.VO2_I.rolling(20, win_type='triang', center=True).mean().fillna(method='bfill').fillna(
        method='ffill')
    df = df.reset_index()
    df = df.drop('timestamp', axis=1)

    if n_inputs == 6:
        # filter_vars = ['VO2_I', 'VCO2_I', 'VE_I', 'HR_I', 'RF_I', 'PetO2_I', 'PetCO2_I']
        if 'VCO2VO2_I' not in df.columns:
            df['VCO2VO2_I'] = df['VCO2_I'].values/df['VO2_I'].values
        filter_vars = ['VCO2VO2_I', 'VE_I', 'PetO2_I', 'PetCO2_I', 'VEVO2_I', 'VEVCO2_I']
        X = df[filter_vars]
        # XN = normalize(X)
        XN = X
        XN['VCO2VO2_I'] = (XN['VCO2VO2_I'] - XN['VCO2VO2_I'].min()) / (
                XN['VCO2VO2_I'].max() - XN['VCO2VO2_I'].min())
        XN['VE_I'] = (XN['VE_I'] - XN['VE_I'].min()) / (
                XN['VE_I'].max() - XN['VE_I'].min())
        XN['PetO2_I'] = (XN['PetO2_I'] - XN['PetO2_I'].min()) / (
                XN['PetO2_I'].max() - XN['PetO2_I'].min())
        # in the case of PetCO2 you take MAX from top 25 and MIN from the single test
        # (x - MIN) / (MAX - MIN)
        XN['PetCO2_I'] = (XN['PetCO2_I'] - XN['PetCO2_I'].min()) / (
                XN['PetCO2_I'].max() - XN['PetCO2_I'].min())
        XN['VEVO2_I'] = (XN['VEVO2_I'] - XN['VEVO2_I'].min()) / (
                XN['VEVO2_I'].max() - XN['VEVO2_I'].min())
        XN['VEVCO2_I'] = (XN['VEVCO2_I'] - XN['VEVCO2_I'].min()) / (
                XN['VEVCO2_I'].max() - XN['VEVCO2_I'].min())
        XN = XN.filter(filter_vars, axis=1)

    if n_inputs == 5 and past_points == 40:
        # TODO: FIX the normalisation here as well
        filter_vars = ['VO2_I', 'VE_I', 'PetO2_I', 'RF_I', 'VEVO2_I']
        X = df[filter_vars]
        XN = normalize(X)
        XN = XN.filter(filter_vars, axis=1)

    p_1 = []
    p_2 = []
    p_3 = []
    time = []

    for i in np.arange(1, len(XN) - int(past_points)):
        XN_array = np.asarray(XN[(i):(i+int(past_points))])
        output_data = tf_model(XN_array.reshape(1, past_points, 6))
        p_1.append(output_data.numpy()[0][0])
        p_2.append(output_data.numpy()[0][1])
        p_3.append(output_data.numpy()[0][2])
        time.append(df.time[i])

    tmp_df = pd.DataFrame()
    tmp_df['time'] = time
    tmp_df['p_md'] = optimal_filter(np.asarray(time), np.asarray(p_1), 100)
    tmp_df['p_hv'] = optimal_filter(np.asarray(time), np.asarray(p_2), 100)
    tmp_df['p_sv'] = optimal_filter(np.asarray(time), np.asarray(p_3), 100)

    # compute the normalised probabilities
    tmp_df['p_md_N'] = np.asarray(p_1) / (np.asarray(p_1) + np.asarray(p_2) + np.asarray(p_3))
    tmp_df['p_hv_N'] = np.asarray(p_2) / (np.asarray(p_1) + np.asarray(p_2) + np.asarray(p_3))
    tmp_df['p_sv_N'] = np.asarray(p_3) / (np.asarray(p_1) + np.asarray(p_2) + np.asarray(p_3))

    # Perform kernel density estimation on the overlap probabilities
    kde_1 = stats.gaussian_kde(time, weights=tmp_df['p_sv_N'] * tmp_df['p_md_N'])
    kde_2 = stats.gaussian_kde(time, weights=tmp_df['p_md_N'] * tmp_df['p_hv_N'])

    overlap_ci_1 = [int(np.percentile(kde_1.resample(4000)[0], 25)), int(np.percentile(kde_1.resample(4000)[0], 75))]
    overlap_ci_2 = [int(np.percentile(kde_2.resample(4000)[0], 25)), int(np.percentile(kde_2.resample(4000)[0], 75))]

    mod_col = tmp_df[['p_md', 'p_hv', 'p_sv']].iloc[:20].mean().idxmax()
    sev_col = tmp_df[['p_md', 'p_hv', 'p_sv']].iloc[-20:].mean().idxmax()
    for labels_ in ['p_md', 'p_hv', 'p_sv']:
        if labels_ not in [mod_col, sev_col]:
            hv_col = labels_

    out_df = pd.DataFrame()
    out_df['time'] = time
    out_df['p_md'] = tmp_df[mod_col]
    out_df['p_hv'] = tmp_df[hv_col]
    out_df['p_sv'] = tmp_df[sev_col]

    if plot == True:
        plot([out_df['p_md'], out_df['p_hv'], out_df['p_sv']],
             title="Exercise intensity domains",
             width=120,
             color=True,
             legend_labels=['1', '2', '3'])

    out_dict = {}
    out_dict['VT1'] = {}
    out_dict['VT2'] = {}
    out_dict['VT1']['time'] = {}
    out_dict['VT2']['time'] = {}
    out_dict['VT1_upper'] = {}
    out_dict['VT1_lower'] = {}
    out_dict['VT2_upper'] = {}
    out_dict['VT2_lower'] = {}

    # FIXME: hard coded
    VT1_index = int(out_df[(out_df['p_hv'] <= out_df['p_md'])].index[-1])
    VT2_index = int(out_df[(out_df['p_hv'] >= out_df['p_sv']) & (out_df['p_hv'] > out_df['p_md'])].index[-1])

    out_dict['VT1']['time'] = df.iloc[VT1_index]['time']
    out_dict['VT2']['time'] = df.iloc[VT2_index]['time']

    out_dict['VT1_upper']['time'] = overlap_ci_1[1]
    out_dict['VT1_lower']['time'] = overlap_ci_1[0]
    out_dict['VT2_upper']['time'] = overlap_ci_2[1]
    out_dict['VT2_lower']['time'] = overlap_ci_2[0]

    out_dict['VT1']['HR'] = df.iloc[VT1_index]['HR_I']
    out_dict['VT2']['HR'] = df.iloc[VT2_index]['HR_I']
    out_dict['VT1_upper']['HR'] = df[df.time == overlap_ci_1[1]]['HR_I'].values[0]
    out_dict['VT1_lower']['HR'] = df[df.time == overlap_ci_1[0]]['HR_I'].values[0]
    out_dict['VT2_upper']['HR'] = df[df.time == overlap_ci_2[1]]['HR_I'].values[0]
    out_dict['VT2_lower']['HR'] = df[df.time == overlap_ci_2[0]]['HR_I'].values[0]

    out_dict['VT1']['VE'] = df.iloc[VT1_index]['VE_I']
    out_dict['VT2']['VE'] = df.iloc[VT2_index]['VE_I']
    out_dict['VT1_upper']['VE'] = df[df.time == overlap_ci_1[1]]['VE_I'].values[0]
    out_dict['VT1_lower']['VE'] = df[df.time == overlap_ci_1[0]]['VE_I'].values[0]
    out_dict['VT2_upper']['VE'] = df[df.time == overlap_ci_2[1]]['VE_I'].values[0]
    out_dict['VT2_lower']['VE'] = df[df.time == overlap_ci_2[0]]['VE_I'].values[0]

    out_dict['VT1']['VO2'] = df.iloc[VT1_index]['VO2_20s']
    out_dict['VT2']['VO2'] = df.iloc[VT2_index]['VO2_20s']
    out_dict['VT1_upper']['VO2'] = df[df.time == overlap_ci_1[1]]['VO2_20s'].values[0]
    out_dict['VT1_lower']['VO2'] = df[df.time == overlap_ci_1[0]]['VO2_20s'].values[0]
    out_dict['VT2_upper']['VO2'] = df[df.time == overlap_ci_2[1]]['VO2_20s'].values[0]
    out_dict['VT2_lower']['VO2'] = df[df.time == overlap_ci_2[0]]['VO2_20s'].values[0]

    return out_df, out_dict

def create_probabilities(duration=600, VT1=320, VT2=460, training=True, normalization=False, resting = True, resting_duration=60, y_pm0=1):
    """Creates the probabilities of being in different intensity domains

    These probabilities are then sent to the CPET generator and they are used ot generate CPET vars that can replicate those probabilities

    Parameters:
        duration (int): Length of the test file
        VT1 (int): First ventilatory threshold, in time samples from the beginning of the test
        VT2 (int): Second ventilatory threshold, in time samples from the beginning of the test
        y_pm0 (double): This is used to set the first value of the probabilities. Indeed if you start from a higher VO2 you're not likely full in moderate domain (pm<1)

    Returns:
        p_mF (np array): Probability of being in the moderate intensity zone (-1:1)
        p_hF (np array): Probability of being in the heavy intensity zone (-1:1)
        p_sF (np array): Probability of being in the severe intensity zone (-1:1)

    """

    import numpy as np
    from scipy.interpolate import interp1d

    t = np.arange(1, duration + 1)

    if resting == True:
        step_1 = resting_duration
        smooth_lambda = [duration*2, duration*2, duration*2]
    else:
        # no resting is included
        step_1 = 1
        smooth_lambda = [duration*2, duration*2, duration*2]

    if y_pm0 < 0.65:
        print('This is too close to VT1, I am settling at 0.65')
        y_pm0 = 0.65

    y_pm = [y_pm0, y_pm0, 0.5, 0, 0, 0]
    y_ph = [1 - y_pm0, 1 - y_pm0, 0.5, 1, 0.5, 0]
    y_ps = [0, 0, 0, 0, 0.5, 1]
    # linear coefficients
    x_p = [0, step_1, VT1, ((VT2 - VT1) / 2 + VT1), VT2, duration]

    from scipy.interpolate import UnivariateSpline
    # build splines
    # moderate => 2 splines
    pm_L1 = UnivariateSpline([0, step_1, VT1], [y_pm0, y_pm0, 0.5], k=2)
    # compute additional points
    pm_L2 = UnivariateSpline([VT1, VT1 + 30, duration - 60, duration - 30, duration], [0.5, pm_L1(VT1 + 30), 0, 0, 0], k=2)
    p_mF = np.hstack((pm_L1(t[t < VT1]), pm_L2(t[t >= VT1])))
    # heavy => 3 splines
    ph_L1 = UnivariateSpline([0, step_1, VT1], [1 - y_pm0, 1 - y_pm0, 0.5], k=2)
    # compute additional points
    ph_L2 = UnivariateSpline([VT1, VT1 + 30, VT2], [0.5, ph_L1(VT1 + 30), 0.5], k=2)
    ph_L3 = UnivariateSpline([VT2, VT2 + 30, duration], [0.5, ph_L2(VT2 + 30), 0], k=2)
    p_hF = np.hstack((ph_L1(t[t < VT1]), ph_L2(t[(t >= VT1) & (t < VT2)]), ph_L3(t[t >= VT2])))
    # severe => 2 splines
    ps_L1 = UnivariateSpline([0, step_1, VT2], [0, 0, 0.5], k=2)
    # compute additional points
    ps_L2 = UnivariateSpline([VT2, VT2 + 15, VT2 + 30, duration - 30, duration],
                             [0.5, ps_L1(VT2 + 15), ps_L1(VT2 + 30), 1, 1], k=1)
    p_sF = np.hstack((ps_L1(t[t < VT2]), ps_L2(t[t >= VT2])))

    p_mF = optimal_filter(t, p_mF, 100)
    p_hF = optimal_filter(t, p_hF, 50)
    p_sF = optimal_filter(t, p_sF, 100)

    if normalization:
        p_mF = np.interp(p_mF, (p_mF.min(), p_mF.max()), (0, 1))
        p_hF = np.interp(p_hF, (p_hF.min(), p_hF.max()), (0, 1))
        p_sF = np.interp(p_sF, (p_sF.min(), p_sF.max()), (0, 1))
    else:
        pass

    if training:
        p_mF = p_mF + np.random.randn(len(t)) / 18
        p_hF = p_hF + np.random.randn(len(t)) / 18
        p_sF = p_sF + np.random.randn(len(t)) / 18
    else:
        pass

    return p_mF, p_hF, p_sF

def random_walk(length=1, scale_factor=1, variation=1):
    """Random walk generator

    Parameters:
        length (int): Length of the output list
        scale_factor (float): Scale factor to be applied to the whole output
        variation (float): Local variation of the main signal with the random walk

    Returns:
        none

    """

    from random import seed
    from random import random

    random_walk = list()
    random_walk.append(-variation if random() < 0.5 else variation)
    for i in range(1, length):
        movement = -variation if random() < 0.5 else variation
        value = random_walk[i - 1] + movement
        random_walk.append(value)

    return [i/scale_factor for i in random_walk]

def generate_CPET(generator,
                  plot=False,
                  fitness_group=None,
                  VT1=None, VT2=None,
                  duration=None,
                  noise_factor=0,
                  resting=False,
                  training=True,
                  normalization=False):
    """Actually generates the CPET file

    Parameters:
        length (int): Length of the output list
        fitness_group (int): Fitness level: low (1), medium (2), high (3). Default to random.
        noise_factor (float): Noise factor for white noise. Default to random.

    Returns:
        df (pd df): Pandas dataframe with CPET data included and ready to be processed by the model (if needed)
        data (dict): Data relative to the generated CPET

    """

    import random
    import numpy as np
    import pandas as pd
    from uniplot import plot as terminal_plot
    from datetime import datetime

    import pkgutil
    from io import StringIO

    if resting:
        bytes_data = pkgutil.get_data('pyoxynet.data_test', 'database_statistics_resting.csv')
    else:
        bytes_data = pkgutil.get_data('pyoxynet.data_test', 'database_statistics_ramp.csv')

    s = str(bytes_data, 'utf-8')
    data = StringIO(s)
    db_df = pd.read_csv(data)

    bytes_data_standard = pkgutil.get_data('pyoxynet.data_test', 'database_statistics_resting.csv')
    s = str(bytes_data_standard, 'utf-8')
    data = StringIO(s)
    db_df_standard = pd.read_csv(data)

    standard_beginning_ratio = db_df_standard.VO2min.mean()/db_df_standard.VO2peak.mean()

    # extract sample from db
    if fitness_group == None:
        # if fitness group is not user defined, then a sample is randomly taken
        db_df_sample = db_df.sample()
    else:
        db_df_sample = db_df[db_df['fitness_group'] == fitness_group].sample()

    if duration == None:
        duration = int(db_df_sample.duration)

    if VT1 == None and VT2 == None:
        VT1 = 0
        VT2 = 0
        # check if difference in threshold is < 10% of the entire duration
        while (duration - VT2) < 60 or (VT2 - VT1) > 240 or (VT2 - VT1) < 60 or (duration - VT2) > 240 or (VT2 - VT1)/duration < 0.2:
            VT1 = int(random.randrange(25, 65) * duration * 0.01)
            VT2 = int(random.randrange(70, 96) * duration * 0.01)

    VO2_peak = int(db_df_sample.VO2peak)
    VCO2_peak = int(db_df_sample.VCO2peak)
    VE_peak = int(db_df_sample.VEpeak)
    RF_peak = int(db_df_sample.RFpeak)
    HR_peak = int(db_df_sample.HRpeak)
    PetO2_peak = int(db_df_sample.PetO2peak)
    PetCO2_peak = int(db_df_sample.PetCO2peak)

    VO2_min = int(db_df_sample.VO2min)
    VCO2_min = int(db_df_sample.VCO2min)
    VE_min = int(db_df_sample.VEmin)
    RF_min = int(db_df_sample.RFmin)
    HR_min = int(db_df_sample.HRmin)
    PetO2_min = int(db_df_sample.PetO2min)
    PetCO2_min = int(db_df_sample.PetCO2min)

    # probability definition
    # FIXME: hard coded here

    tmp_beginning_ratio = VO2_min/VO2_peak
    y_pm0 = np.min([standard_beginning_ratio/tmp_beginning_ratio, 1])

    p_mF, p_hF, p_sF = create_probabilities(duration=duration,
                                            VT1=VT1,
                                            VT2=VT2,
                                            training=training,
                                            resting=resting,
                                            normalization=normalization,
                                            y_pm0=y_pm0)

    # initialise
    VO2 = []
    VCO2 = []
    VE = []
    HR = []
    RF = []
    PetO2 = []
    PetCO2 = []

    time_array = np.arange(duration)
    # TODO: this is hard coded
    input_data = np.array(np.random.random_sample([1, 53]))

    for seconds_ in time_array:
        # keep the seed?
        # input_data = np.array(np.random.random_sample(input_shape), dtype=np.float32)
        input_data[0, -3:] = np.array([[p_hF[seconds_], p_sF[seconds_], p_mF[seconds_]]])
        output_data = generator(input_data)
        VO2.append(np.median(output_data[0, :, 0]))
        VCO2.append(np.median(output_data[0, :, 1]))
        VE.append(np.median(output_data[0, :, 2]))
        HR.append(np.median(output_data[0, :, 3]))
        RF.append(np.median(output_data[0, :, 4]))
        PetO2.append(np.median(output_data[0, :, 5]))
        PetCO2.append(np.median(output_data[0, :, 6]))
        # vars -> ['VO2_I', 'VCO2_I', 'VE_I', 'HR_I', 'RF_I', 'PetO2_I', 'PetCO2_I']

    # filter before you expand again between min and max
    VO2 = optimal_filter(time_array, VO2, 10)
    VCO2 = optimal_filter(time_array, VCO2, 10)
    VE = optimal_filter(time_array, VE, 10)
    HR = optimal_filter(time_array, HR, 10)
    RF = optimal_filter(time_array, RF, 10)
    PetO2 = optimal_filter(time_array, PetO2, 10)
    PetCO2 = optimal_filter(time_array, PetCO2, 10)

    min_norm = 0
    max_norm = 1
    VO2 = np.interp(np.asarray(VO2), (np.asarray(VO2).min(), np.asarray(VO2).max()), (min_norm, max_norm))
    VCO2 = np.interp(np.asarray(VCO2), (np.asarray(VCO2).min(), np.asarray(VCO2).max()), (min_norm, max_norm))
    VE = np.interp(np.asarray(VE), (np.asarray(VE).min(), np.asarray(VE).max()), (min_norm, max_norm))
    HR = np.interp(np.asarray(HR), (np.asarray(HR).min(), np.asarray(HR).max()), (min_norm, max_norm))
    RF = np.interp(np.asarray(RF), (np.asarray(RF).min(), np.asarray(RF).max()),
                   (min_norm, max_norm))
    PetO2 = np.interp(np.asarray(PetO2), (np.asarray(PetO2).min(), np.asarray(PetO2).max()), (min_norm, max_norm))
    PetCO2 = np.interp(np.asarray(PetCO2), (np.asarray(PetCO2).min(), np.asarray(PetCO2).max()),
                       (min_norm, max_norm))

    df = pd.DataFrame()
    df['time'] = time_array

    if noise_factor == None:
        noise_factor = random.randint(3, 5)/2
    else:
        pass

    # TODO: should I compute VEVO2 or VEVCO2 before adding the noise (?)
    df['VO2_I'] = (np.asarray(VO2) - np.min(VO2))/(np.max((np.asarray(VO2) - np.min(VO2)))) * (VO2_peak - VO2_min) + VO2_min + np.random.randn(len(VO2)) * 40 * noise_factor
    df['VCO2_I'] = (np.asarray(VCO2) - np.min(VCO2))/(np.max((np.asarray(VCO2) - np.min(VCO2)))) * (VCO2_peak - VCO2_min) + VCO2_min + np.random.randn(len(VO2)) * 40 * noise_factor
    df['VE_I'] = (np.asarray(VE) - np.min(VE))/(np.max((np.asarray(VE) - np.min(VE)))) * (VE_peak - VE_min) + VE_min + np.random.randn(len(VO2)) * 1.5 * noise_factor
    df['HR_I'] = np.ndarray.astype((np.asarray(HR) - np.min(HR)) / (np.max((np.asarray(HR) - np.min(HR)))) * (HR_peak - HR_min) + HR_min + np.random.randn(len(VO2)) * 1 * noise_factor, int)
    df['RF_I'] = (np.asarray(RF) - np.min(RF)) / (np.max((np.asarray(RF) - np.min(RF)))) * (
            RF_peak - RF_min) + RF_min + np.random.randn(len(VO2)) * 1 * noise_factor
    df['PetO2_I'] = (np.asarray(PetO2) - np.min(PetO2))/(np.max((np.asarray(PetO2) - np.min(PetO2)))) * (PetO2_peak - PetO2_min) + PetO2_min + np.random.randn(len(VO2)) * 1.5 * noise_factor
    df['PetCO2_I'] = (np.asarray(PetCO2) - np.min(PetCO2))/(np.max((np.asarray(PetCO2) - np.min(PetCO2)))) * (PetCO2_peak - PetCO2_min) + PetCO2_min + np.random.randn(len(VO2)) * 1.5 * noise_factor

    df['p_mF'] = p_mF
    df['p_hF'] = p_hF
    df['p_sF'] = p_sF

    df['VEVO2_I'] = df['VE_I']/df['VO2_I']
    df['VEVCO2_I'] = df['VE_I']/df['VCO2_I']

    df['VCO2VO2_I'] = df['VCO2_I'] / df['VO2_I']
    df['PetO2VO2_I'] = df['PetO2_I'] / df['VO2_I']
    df['PetCO2VO2_I'] = df['PetCO2_I'] / df['VO2_I']

    df['domain'] = np.NaN
    df.loc[df['time'] < (VT1), 'domain'] = -1
    df.loc[df['time'] >= (VT2), 'domain'] = 1
    df.loc[(df['time'] < (VT2)) & (df['time'] >= (VT1)), 'domain'] = 0
    df['fitness_group'] = db_df_sample['fitness_group'].values[0]
    df['Age'] = db_df_sample['Age'].values[0]
    df['age_group'] = db_df_sample['age_group'].values[0]
    df['gender'] = db_df_sample['gender'].values[0]
    df['weight'] = db_df_sample['weight'].values[0]
    df['height'] = db_df_sample['height'].values[0]

    # Collect VO2 value at VT1 and VT2
    VO2VT1 = df.iloc[(df[df['domain'].diff().fillna(0) == 1].index[0] - 10):(df[df['domain'].diff().fillna(0) == 1].index[0] + 10)]['VO2_I'].mean()
    VO2VT2 = df.iloc[(df[df['domain'].diff().fillna(0) == 1].index[1] - 10):(
            df[df['domain'].diff().fillna(0) == 1].index[1] + 10)]['VO2_I'].mean()

    if plot:
        terminal_plot([df['VO2_I'], df['VCO2_I']],
                      title="CPET variables", width=120,
                      color=True, legend_labels=['VO2_I', 'VCO2_I'])
        terminal_plot([df['VE_I']],
                      title="CPET variables", width=120,
                      color=True, legend_labels=['VE'])
        terminal_plot([df['PetO2_I'], df['PetCO2_I']],
                      title="CPET variables", width=120,
                      color=True, legend_labels=['PetO2_I', 'PetCO2_I'])

    if db_df_sample.fitness_group.values == 1:
        fitness_group = 'LOW'
    if db_df_sample.fitness_group.values == 2:
        fitness_group = 'MEDIUM'
    if db_df_sample.fitness_group.values == 3:
        fitness_group = 'HIGH'
    if db_df_sample.gender.values == -1:
        gender = 'MALE'
    if db_df_sample.gender.values == 1:
        gender = 'FEMALE'

    print('Data generated for a ', gender, ' individual with ', fitness_group, ' fitness capacity.')
    print('Weight: ', int(db_df_sample.weight.values), ' kg')
    print('Height: ', db_df_sample.height.values[0], 'm')
    print('Age: ', int(db_df_sample.Age.values), 'y')
    print('Noise factor: ', round(noise_factor, 2))
    print('VT1: ', str(VT1))
    print('VT2: ', str(VT2))
    print('VO2VT2: ', str(int(VO2VT1)))
    print('VO2VT2: ', str(int(VO2VT2)))
    print('Resting:', resting)

    # TODO: create a function to generate this dict, as it is the same that we should use when drawing real data files
    # create dict fro Exercise Threshold App
    data = dict()
    data['Age'] = str(int(db_df_sample.Age.values))
    data['Height'] = str(db_df_sample.height.values[0])
    data['Weight'] = str(int(db_df_sample.weight.values))
    data['Gender'] = gender
    data['Aerobic_fitness_level'] = fitness_group
    data['VT1'] = str(VT1)
    data['VT2'] = str(VT2)
    data['VO2VT1'] = str(int(VO2VT1))
    data['VO2VT2'] = str(int(VO2VT2))
    data['VO2max'] = str(int(VO2_peak))
    data['LT'] = str(int(VO2VT1))
    data['LT_vo2max'] = str(int((VO2VT1/VO2_peak)*100)) + '%'
    data['RCP'] = str(int(VO2VT2))
    data['RCP_vo2max'] = str(int((VO2VT2/VO2_peak) * 100)) + '%'
    data['id'] = 'fake_#'
    data['noise_factor'] = str(noise_factor)
    data['created'] = datetime.today().strftime("%m/%d/%Y - %H:%M:%S")
    data['resting'] = str(resting)

    df['breaths'] = (1/(df.RF_I/60))
    # df['recorded_timestamp'] = pd.to_datetime(df.time, unit='s')

    time_cum_sum = 0
    df_breath = pd.DataFrame()
    n = 0
    while time_cum_sum < (duration-6):
        try:
            tmp = df[(df.time >= time_cum_sum) & (df.time < (time_cum_sum + df.breaths.iloc[n]))].median()
            # TODO: this 25 is hardcoded, you should have len(df.columns)
            df_breath = pd.concat([df_breath, pd.DataFrame(data=np.reshape(tmp.values, [1, 24]),
                                                           columns=tmp.index.to_list())])
            time_cum_sum = (time_cum_sum + df.breaths.iloc[n])
            n = n + len(df[(df.time >= time_cum_sum) & (df.time < (time_cum_sum + df.breaths.iloc[n]))])
        except:
            break

    df_breath = df_breath.dropna()
    df_breath['time'] = df_breath['time'].astype(int)
    df_breath = df_breath.drop_duplicates('time')

    exercise_threshold_names = {"time": "t",
                                "VO2_I": "VO2",
                                "VCO2_I": "VCO2",
                                "VE_I": "VE",
                                "VCO2VO2_I": "R",
                                "VEVO2_I": "VE/VO2",
                                "VEVCO2_I": "VE/VCO2",
                                "PetO2_I": "PetO2",
                                "PetCO2_I": "PetCO2",
                                "HR_I": "HR",
                                "RF_I": "RF"}

    df_breath = df_breath.rename(columns=exercise_threshold_names)
    # create the dict for the exercise threshold app
    data['data'] = dict()
    data['data'] = df_breath.to_dict(orient='records')

    return df, data